#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <stdexcept>

#include "GTXLReader.h"
#include "rule.h"
#include "hypergraph.h"
#include <string.h>

namespace algorithms {

namespace readers {

using namespace data_structures;


GTXLREADER::GTXLREADER()
{
	name = "GtxlReader";
}

GTXLREADER::~GTXLREADER()
{
	xmlFreeDoc(doc);
}  
    
void GTXLREADER::initialize(char *filename)
{

    doc = xmlParseFile(filename);
	
    if (doc == NULL ) 
      {
	fprintf(stderr,"Document not parsed successfully. \n");
	xmlFreeDoc(doc);
	cout << "XML parser error\n";
	exit(1);
      }
	
	
    if (xmlStrcmp(xmlDocGetRootElement(doc)->name, (const xmlChar *) "GTS")) 
      {
	fprintf(stderr,"document of the wrong type, root node != GTS");
	xmlFreeDoc(doc);
	exit(1);
      }
}
    

void GTXLREADER::run(GTS &GraphGrammar)
{

xmlNodePtr root = xmlDocGetRootElement(doc);
if(xmlStrcmp(root->name,(const xmlChar *)"GTS"))
{
	cout << "Root node 'GTS' expected \n";
	exit(1);
}
xmlNodePtr in_graph = GetChildrenNode(root,"Initial");
if(in_graph == NULL)
{
	cout << "Initial file not found\n";
	exit(1);
}

HYPERGRAPH hg;

//RULE *rule = new RULE();

//variables to check id's unique
    set<string> rule_ids;
    pair<set<string>::iterator, bool> rules_checker;
    string id_string;
    set<string> ids;
    
    hg = ReadHypergraph(in_graph, ids, 0);

	GraphGrammar.setStartgraph(&hg);
	
	xmlNodePtr rule_graph = GetChildrenNode(root, "Rule");
	if(rule_graph == NULL)
	{
		cout << "Initial file without rules found\n";
		exit(1);
	}
	

IDTYPE id = hg.getLastUsedID();

while(rule_graph != NULL)
{
	RULE rule;
	// clean the array of ids. The ids must be unique inside of one rule
	ids.clear();
	id_string = (string)(char*)xmlGetProp(rule_graph,(const xmlChar *)"id");
	rules_checker = rule_ids.insert(id_string);
	if (rules_checker.second == false)
	{
		printf("Error: The id's in 'Rule' are not unique: id = %s\n", id_string.c_str());
		exit(1);
	} 
	
	rule.setName(id_string);
	
	rule.setID(++id);

	// guard
	xmlNodePtr guard_node = GetChildrenNode(rule_graph, "precondition");
	if(guard_node != NULL)
	{
		xmlNodePtr guard_node2 = GetChildrenNode(guard_node, "condition");
		if(guard_node2 != NULL)
		{
			xmlNodePtr guard_node3 = GetChildrenNode(guard_node2, "attrCondition");
			if(guard_node3 != NULL)
			{
				xmlNodePtr guards_node = GetChildrenNode(guard_node3, "string");
				if(guards_node != NULL) 
				{
					guards_node = guards_node->xmlChildrenNode;
					if (guards_node != NULL)
					{
						string guards = (string)((char*)guards_node->content);

						char ch[guards.size()+1];
						strcpy(ch, guards.c_str());
						char * pch;
						pch = strtok(ch,",");
						
						while (pch!=NULL)
						{
							EXPRESSION eg;
							eg.addName("guard");
							TYPES types;
							types.push_back(BOOL);
							eg.addType(types);
							eg.addExpr((string)pch);
							rule.addGuard(eg);
							pch = strtok (NULL, ",");
						}
					}
				}
			}
		}
	}

	xmlNodePtr left_rule = GetChildrenNode(rule_graph, "LHS");
	if(left_rule == NULL)
	{
		cout << "Rule without left side found\n";
		exit(1);
	}
	
	left_rule = GetChildrenNode(left_rule, "RuleGraph");
	if(left_rule == NULL)
	{
		cout << "Rule without RuleGraph found\n";
		exit(1);
	}
	
	id_string = (string)(char*)xmlGetProp(left_rule,(const xmlChar *)"id");
	rules_checker = ids.insert(id_string);
	if (rules_checker.second == false)
	{
		printf("Error: The id's in 'RuleGraph' are not unique: id = %s\n", id_string.c_str());
		exit(1);
	} 
	
	HYPERGRAPH hg_left = ReadHypergraph(left_rule, ids, id);
	rule.setLeft(&hg_left);
	
	
	//cout << GraphGrammar.getRules()->size()<<endl;
	xmlNodePtr right_rule = GetChildrenNode(rule_graph, "RHS");
	if(right_rule == NULL)
	{
		cout << "Rule without right side found\n";
		exit(1);
	}
	
	right_rule = GetChildrenNode(right_rule, "RuleGraph");
	if(right_rule == NULL)
	{
		cout << "Rule without RuleGraph found\n";
		exit(1);
	}
	
	id_string = (string)(char*)xmlGetProp(right_rule,(const xmlChar *)"id");
	rules_checker = ids.insert(id_string);
	if (rules_checker.second == false)
	{
		printf("Error: The id's in 'RuleGraph' are not unique: id = %s\n", id_string.c_str());
		exit(1);
	} 
		
	IDTYPE lastID = rule.getLeft()->getLastUsedID();
	HYPERGRAPH hg_right = ReadHypergraph(right_rule, ids, lastID);
	rule.setRight(&hg_right);
	
	// read mappings
	xmlNodePtr mapping_xml = GetChildrenNode(rule_graph, "Mapping");
	if(mapping_xml == NULL)
	{
		cout << "Rule without mapping found\n";
		exit(1);
	}
	
	MATCH mapping;

		id_string = (string)(char*)xmlGetProp(mapping_xml,(const xmlChar *)"id");	
		rules_checker = rule_ids.insert(id_string);
		if (rules_checker.second == false)
		{
			printf("Error: The id's in 'Mapping' are not unique: id = %s\n", id_string.c_str());
			exit(1);
		} 
		
		xmlNodePtr map_elem = GetChildrenNode(mapping_xml, "MapElem");
		while(map_elem != NULL)
		{
			string fromXMLID = (string)(char*)xmlGetProp(map_elem,(const xmlChar *)"from");
	    		string toXMLID = (string)(char*)xmlGetProp(map_elem,(const xmlChar *)"to");
			
			// variables to check existence of mapping connections
			bool ch1 = false;
			bool ch2 = false;
			IDTYPE id1, id2;
			
			// find out the type of connections
			if (rule.getLeft()->hasVertex(fromXMLID))
				{
					id1 = rule.getLeft()->getVertex(fromXMLID).getID();
					ch1 = true;
				}
			if (rule.getLeft()->hasEdge(fromXMLID))
				{
					id1 = rule.getLeft()->getEdge(fromXMLID).getID();
					ch1 = true;
				}
			if (rule.getRight()->hasVertex(toXMLID))
				{
					id2 = rule.getRight()->getVertex(toXMLID).getID();
					ch2 = true;
				}
			if (rule.getRight()->hasEdge(toXMLID))
				{
					id2 = rule.getRight()->getEdge(toXMLID).getID();
					ch2 = true;
				}
			
			if(!(ch1&&ch2))
			{
				cout << "The nodes (edges) described in mapping do not exist in this graph\n";
				cout << "IDS: " << fromXMLID << " " << toXMLID << endl;
				exit(1);
			}
			
			mapping.setMatch(id1, id2);
			map_elem = GetNextChild(map_elem, "MapElem");
		}
	
	rule.setInterface(mapping);
	GraphGrammar.addRule(&rule);
	//delete rule;
	rule_graph = GetNextChild(rule_graph, "Rule");

}


}
    
     
HYPERGRAPH  GTXLREADER::ReadHypergraph(xmlNodePtr root, set<string> &ids, IDTYPE lastID)
{

    xmlNodePtr graph = GetChildrenNode(root, "Graph");
    if (graph == NULL)	
    {
    	cout << "node 'Graph' expected\n";
	exit(1);
    }
	
    // true if hypergraph=true is set
    bool isHypergraph(!xmlStrcmp(xmlGetProp(graph,(const xmlChar *)"hypergraph"),(const xmlChar *)"true"));
    const xmlChar *edgemodeName(xmlGetProp(graph,(const xmlChar *)"edgemode"));
    
    if( isHypergraph && (!xmlStrcmp(edgemodeName,(const xmlChar *)"directed")))
      {
	cerr << "This program does not support directed hypergraphs.\n";
	cerr << "directed hypergraph in input file.\n";
	exit(1);
      }

    bool isDirected = false;
    if(!xmlStrcmp(edgemodeName,(const xmlChar *)"directed")) 
      isDirected = true;
    if(!xmlStrcmp(edgemodeName,(const xmlChar *)"defaultdirected"))
      isDirected = true;
      
    HYPERGRAPH result;
    result.setLastUsedID(lastID);
    //variables to check id's unique
   // set<string> ids;
    pair<set<string>::iterator, bool> checker;
    string id_str = "";
    
	xmlChar* tXMLID = xmlGetProp(graph, (const xmlChar *)"id");
        if(NULL != tXMLID) id_str = string((char *)tXMLID);
	checker = ids.insert(id_str);
		if (checker.second == false)
		{
			printf("Error: The id's in 'Graph' are not unique: id = %s\n", id_str.c_str());
			exit(1);
		} 

	result.setName(id_str);
	IDTYPE id;
	result.setID();
		
    // start parsing the node/edge list:
    
    //read Nodes
    
    xmlNodePtr graph_element = GetChildrenNode(graph, "node");
    while(graph_element != NULL)
       {
	  //check node's ids
	        id_str = "";
	        tXMLID = xmlGetProp(graph_element, (const xmlChar *)"id");
                if(NULL != tXMLID) id_str = string((char *)tXMLID);
	  	checker = ids.insert(id_str);
		if (checker.second == false)
		{
			printf("Error: The id's in 'node' are not unique: id = %s\n", id_str.c_str());
			exit(1);
		}
		
		VERTEX v;
		v.setIdLabel(id_str);
		id = result.getLastUsedID();
		v.setID(++id);
		result.setLastUsedID(id);
		
		result.addVertex(v,id);
		
		graph_element = GetNextChild(graph_element, "node");
	} 
	  
	
   // read Edges
   list<IDTYPE> list_of_vertices;
   
	if (!isHypergraph)
	{
		// (biconnected) edge
		graph_element = GetChildrenNode(graph, "edge");
	  	while(graph_element != NULL)
       		{
	  		//check edge's IDs
	                id_str = "";
	                tXMLID = xmlGetProp(graph_element, (const xmlChar *)"id");
                        if(NULL != tXMLID) id_str = string((char *)tXMLID);
	  		checker = ids.insert(id_str);
			if (checker.second == false)
			{
				printf("Error: The id's in 'edge' are not unique: id = %s\n", id_str.c_str());
				exit(1);
			}
		
			EDGE edge;	
			
			edge.setIdLabel(id_str);
			id = result.getLastUsedID();
			edge.setID(++id);
			result.setLastUsedID(id);			
			
			string fromXMLID = (string)(char*)xmlGetProp(graph_element,(const xmlChar *)"from");
	    		string toXMLID = (string)(char*)xmlGetProp(graph_element,(const xmlChar *)"to");
			
			if (!((result.hasVertex(fromXMLID))&&(result.hasVertex(toXMLID))))
			{
				cout << "The vertex used in edge's connection does not exist in this graph\n";
				exit(1);
			}

			VERTEX ver1;
			VERTEX ver2;
			list_of_vertices.clear();
			ver1 = result.getVertex(fromXMLID);
			list_of_vertices.push_back(ver1.getID());
			ver2 = result.getVertex(toXMLID);
			list_of_vertices.push_back(ver2.getID());
			
			// add Edge to ConnectedEdges
			result.addEdgeToCE(ver1.getID(), edge.getID());
			result.addEdgeToCE(ver2.getID(), edge.getID());
			
			//conVertices.insert(type_conVertices(id, list_of_vertices));
			
			xmlNodePtr attribut = GetChildrenNode(graph_element, "attr");
			if(attribut != NULL)
				if(!xmlStrcmp(xmlGetProp(attribut,(const xmlChar *)"name"),(const xmlChar *)"label")) 
				{
					attribut = GetChildrenNode(attribut, "string");
					if(attribut != NULL) 
						attribut = attribut->xmlChildrenNode;
						if (attribut != NULL)
						{
							edge.setLabel((char*)attribut->content);
						}
				}	
					
			result.addEdge(edge,id,list_of_vertices); 
			
			graph_element = GetNextChild(graph_element, "edge");
		}
	}
	
	graph_element = GetChildrenNode(graph, "rel");
	if((!isHypergraph)&&(graph_element != NULL)) // only hypergraphs have rel edges
	      {
		cerr << "Hyperedge found in a non-hypergraph\n";
		cerr << "Hyperedge found in a non-hypergraph\n";
		exit(1);
	      }
	      
	while(graph_element != NULL)
	{
		//check edge's IDs
	        id_str = "";
	        tXMLID = xmlGetProp(graph_element, (const xmlChar *)"id");
                if(NULL != tXMLID) id_str = string((char *)tXMLID);
	  	checker = ids.insert(id_str);
		if (checker.second == false)
			{
				printf("Error: The id's in 'edge' are not unique: id = %s\n", id_str.c_str());
				exit(1);
			}
		
		EDGE edge;	
			
		edge.setIdLabel(id_str);
		id = result.getLastUsedID();
		edge.setID(++id);	
		result.setLastUsedID(id);
		
		xmlNodePtr attribut = GetChildrenNode(graph_element, "attr");
		while(attribut != NULL)
		{
			if(!xmlStrcmp(xmlGetProp(attribut,(const xmlChar *)"name"),(const xmlChar *)"label")) 
			{
				xmlNodePtr attribut_label;
				attribut_label = GetChildrenNode(attribut, "string");
				if(attribut_label != NULL) 
					attribut_label = attribut_label->xmlChildrenNode;
					if (attribut_label != NULL)
						edge.setLabel((char*)attribut_label->content);
				attribut = GetNextChild(attribut, "attr");
				continue;
			}
			
			xmlChar* sattr  = xmlGetProp(attribut,(const xmlChar *)"kind");
			if(sattr==NULL)
				continue;

			EXPRESSION e;
			TYPES t;

			char * ch;
			ch = strtok((char*)sattr, "-");

			while(ch!=NULL)
			{
				if(strcmp(ch, "int")==0)
					t.push_back(INTEGER);
				else if(strcmp(ch, "str")==0)
					t.push_back(STRING);
				else if(strcmp(ch, "bool")==0)
					t.push_back(BOOL);

				ch = strtok(NULL, "-");
			}

			e.addType(t);
			xmlChar* sname = xmlGetProp(attribut,(const xmlChar *)"name");
			e.addName((string)(char*)sname);
			xmlNodePtr attribut_attr;
			attribut_attr = GetChildrenNode(attribut, "string");
			if(attribut_attr != NULL) 
				attribut_attr = attribut_attr->xmlChildrenNode;
				if (attribut_attr != NULL)
					e.addExpr((string)(char*)attribut_attr->content);
		
			edge.setType(t);
			edge.addAttributes(e);
			
			attribut = GetNextChild(attribut, "attr");	
		}
		
		int order = 0;
		xmlNodePtr relend = GetRelend(graph_element, order);
		list_of_vertices.clear();
		string s;
		while(relend != NULL)
		{
			s = (string)(char*)xmlGetProp(relend,(const xmlChar *)"target");
			if(!result.hasVertex(s))
			{
				cout << "The vertex (" << s << ") used in edge's connection (relend) does not exist in this graph\n";
				exit(1);
			}
			VERTEX ver;
			ver = result.getVertex(s);
			list_of_vertices.push_back(ver.getID());
			order++;
			relend = GetRelend(graph_element, order);
			
		}
		
		result.addEdge(edge, id, list_of_vertices);
		
		graph_element = GetNextChild(graph_element, "rel");
		
	}

	  return result;

}


xmlNodePtr GTXLREADER:: GetChildrenNode(xmlNodePtr parent, char* name_child)
{
	xmlNodePtr child = parent->xmlChildrenNode;
	while((child != NULL) && (xmlStrcasecmp(child->name, (const xmlChar*)name_child)))
	{
		child = child->next;
	}
	
	return child;
	
		
}

xmlNodePtr GTXLREADER:: GetRelend(xmlNodePtr rel, int order)
{
	string s;
	xmlNodePtr relend = GetChildrenNode(rel, "relend");
	while(relend != NULL)
	{
		s = (char*)xmlGetProp(relend,(const xmlChar *)"startorder");
		if(atoi(s.c_str()) == order)
			return relend;
		relend = GetNextChild(relend, "relend");
	}
	
	return NULL;
}


xmlNodePtr GTXLREADER::GetNextChild(xmlNodePtr child1, char* name_child)
{
	
	xmlNodePtr child2 = child1->next;
	while((child2 != NULL) && (xmlStrcasecmp(child2->name, (const xmlChar*)name_child)))
	{
		child2 = child2->next;
	}
	
	return child2;
	
	
}

  /** Create a new unique ID for vertices, edges, transformation rules etc */
/*IDTYPE GTXLREADER::createNewID()
  {
    lastUsedID++;
    /** detect overflow*/
    /*assert(0 < lastUsedID);
    return lastUsedID;
  }*/


};
};


