/***************************************************************************
 *   Copyright (C) 2007 by Salil Joshi   *
 *   hellfeuer@gmail.com   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "cross_product.h"

namespace algorithms {

namespace hypergraph {

CROSS_PRODUCT::CROSS_PRODUCT()
{
}

void CROSS_PRODUCT::initialize(FINITE_AUTOMATON* f, HYPERGRAPH* h)
{
	fa = f;

	// copy
	hg = new HYPERGRAPH();
	*hg = *h;

	// make arity one edges from arity zero edges
	IDTYPE counter = 0;
	map<IDTYPE, VERTEX> v = hg->getVertices();
	map<IDTYPE, EDGE> e = hg->getEdges();
	
	// initialize counter
	for(map<IDTYPE, VERTEX>::iterator itv = v.begin(); itv != v.end(); itv++)
		if(itv->first > counter)
			counter = itv->first;

	for(map<IDTYPE, EDGE>::iterator ite = e.begin(); ite != e.end(); ite++)
		if(ite->first > counter)
			counter = ite->first;

	// replace edges with arity zero
	for(map<IDTYPE, EDGE>::iterator itz = e.begin(); itz != e.end(); itz++)
	{
		list<IDTYPE> lv = hg->getConnectedVertices(itz->first);
		if(lv.size() == 0)
		{
			VERTEX v;
			v.setID(counter);
			counter++;

			hg->addConnectedVertex(itz->first, v);
		}
	}
}

void CROSS_PRODUCT::run(FINITE_AUTOMATON* fb)
{
	PRODUCTMAP productmap;
	map<IDTYPE, VERTEX>* v = hg->getVerticesPointer();
	
	set<STATE*>::iterator i;
	map<IDTYPE, VERTEX>::iterator j;
	STATE* s;
	
	//product of states with vertices
	for (i = fa->allStates.begin(); i != fa->allStates.end(); i++)
	{
		for (j = v->begin(); j != v->end(); j++)
		{
			s = fb->addState();
			productmap.insert(pair<IDTYPE, IDTYPE>((*j).first, (*i)->id), s);
		}
	}
	
	//create intial states
	for (i = fa->initialStates.begin(); i != fa->initialStates.end(); i++)
	{
		for (j = v->begin(); j != v->end(); j++)
		{
			pair<pair<IDTYPE, IDTYPE>, STATE*>  * p = productmap.find(pair<IDTYPE, IDTYPE>((*j).first, (*i)->id));
			if(p != NULL)
			 fb->initialStates.insert(p->second);
		}
	}
	
	//create final states
	for (i = fa->finalStates.begin(); i != fa->finalStates.end(); i++)
	{
		for (j = v->begin(); j != v->end(); j++)
		{
			pair<pair<IDTYPE, IDTYPE>, STATE*> * p = productmap.find(pair<IDTYPE, IDTYPE>((*j).first, (*i)->id));
			if(p != NULL)
			 fb->finalStates.insert(p->second);
		}
	}
	
	map<IDTYPE, EDGE> * edges = hg->getEdgesPointer();
	map<IDTYPE, EDGE>::iterator k;

	//transitions
	for (k = edges->begin(); k != edges->end(); k++)
	{
		string label = k->second.getLabel();
		char ch[32];
	    	sprintf(ch, "%i", k->first);
		string label_ID = label + string("_") + string(ch);

		map<pair<STATE*, string>, set<STATE*> >::iterator trans;
		
		for (trans = fa->transitions.begin(); trans != fa->transitions.end(); trans++)
		{
			string transLabel = trans->first.second;
			if (transLabel == label) //only edge label is specified
			{
				list<IDTYPE> connectedV = hg->getConnectedVertices(k->second.getID());
				IDTYPE from = *(connectedV.begin());
				connectedV.pop_front();
				
				// arity > 0
				if(connectedV.size() > 0)
				{
					list<IDTYPE>::iterator connViter;
					for(connViter = connectedV.begin(); connViter != connectedV.end(); connViter++)
					{
						
						createTransition(from, *connViter, trans->first.first, trans->second, label_ID, fb, &productmap);
					}
				}
				// arity == 0
				else
				{
					createTransition(from, from, trans->first.first, trans->second, label_ID, fb, &productmap);
				}
			} else {
				size_t pos = transLabel.find(label); //find this edge label in the transition
				if (pos != string::npos)
				{
					list<IDTYPE> tmpV = hg->getConnectedVertices(k->second.getID());
					vector<IDTYPE> connectedV = vector<IDTYPE>(tmpV.begin(), tmpV.end());//this is temporary, to allow random access. the copy shouldnt be done.
					int fromVertex = atoi(transLabel.substr(0, pos).c_str());
					int toVertex = atoi(transLabel.substr(pos+label.length()).c_str());
					createTransition(connectedV[fromVertex], connectedV[toVertex], trans->first.first, trans->second, label_ID, fb, &productmap);
				}
			}	
		}
	}
}

void CROSS_PRODUCT::createTransition(IDTYPE fromId, IDTYPE toId, STATE* fromState,
	 set<STATE*> &toStates, string label, FINITE_AUTOMATON* fb, PRODUCTMAP * productmap)
{
	pair<pair<IDTYPE, IDTYPE>, STATE*> * p = productmap->find(pair<IDTYPE, IDTYPE>(fromId, fromState->id));
	STATE* from = NULL;
	if(p != NULL)
	 from = p->second;
	else
	{
	 cout << "Internal Error in CROSS_PRODUCT: NULL Pointer" << endl;
	 exit(1);
	}
	
	set<STATE*>::iterator i;
	for (i = toStates.begin(); i != toStates.end(); i++)
	{
		pair<pair<IDTYPE, IDTYPE>, STATE*> * p = productmap->find(pair<IDTYPE, IDTYPE>(toId, (*i)->id));
		if(p != NULL)
		 fb->addTransition(from, label, p->second);
	}
}

CROSS_PRODUCT::~CROSS_PRODUCT()
{
}

void CROSS_PRODUCT::PRODUCTMAP::insert(pair<IDTYPE, IDTYPE> p, STATE * s)
{
	if(find(p) == NULL)
	 prmap.push_back(pair<pair<IDTYPE, IDTYPE>, STATE*>(p,s));
}

pair<pair<IDTYPE, IDTYPE>, STATE*>  * CROSS_PRODUCT::PRODUCTMAP::find(pair<IDTYPE, IDTYPE> p)
{
	for(list<pair<pair<IDTYPE, IDTYPE>, STATE*> >::iterator it = prmap.begin(); it != prmap.end(); it++)
	{
		if(it->first.first == p.first && it->first.second == p.second)
			return &(*it);
	}

	return NULL;
}

}

}
