/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         * 
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef ALGORITHM_GTXL_NAC_READER_CPP
#define ALGORITHM_GTXL_NAC_READER_CPP

#include "gtxl_nac_reader.h"
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>

namespace algorithms
{
   namespace readers
   {
      GTXL_NAC_READER::GTXL_NAC_READER()
      {
	 this->lastUsedID = debug? 1: 0;
      }
      GTXL_NAC_READER::~GTXL_NAC_READER()
      {
      }

      void GTXL_NAC_READER::initialize(char *filename)
      {
	 this->fname = (string) filename;
      }

      void GTXL_NAC_READER::run(GTS & GraphGrammar)
      {
         bool interpretNAC = true;
        
	 /*
	  * initialize graphgrammar
	  */
	 GraphGrammar = GTS();
	 HYPERGRAPH *initialGraph = NULL;
	 map < IDTYPE, RULE > rules;
      
	 /*
	  * open file
	  */
	 xmlDocPtr doc = xmlParseFile(this->fname.c_str());
	 if(doc == NULL)
	 {
	    fprintf(stdout, "Document not parsed successfully. \n");
	    xmlFreeDoc(doc);
	    exit(1);
	 }
	 if(xmlStrcmp(xmlDocGetRootElement(doc)->name, (const xmlChar *)"GTS"))
	 {
	    fprintf(stdout, "document of the wrong type, root node != GTS");
	    xmlFreeDoc(doc);
	    exit(1);
	 }
	 xmlNodePtr root = xmlDocGetRootElement(doc);

	 xmlNodePtr cur = root->xmlChildrenNode;
	 while(NULL != cur)
	 {
	    if(!xmlStrcmp(cur->name, (const xmlChar *)"Initial"))	// initial graph?
	    {
	       if(NULL != initialGraph)	// is there already an initial graph?
	       {
		  cout <<
		     "There is more than one initial graph specified in the input file.\n";
		  exit(1);
	       }
	       else		// first initial graph?
	       {
		  // Initial contains a Graph node
		  xmlNodePtr igraph = cur->xmlChildrenNode;
		  while(NULL != igraph
			&& xmlStrcmp(igraph->name, (const xmlChar *)"Graph"))
		     igraph = igraph->next;
		  if(NULL == igraph)
		  {
		     cout << "Initial without Graph section found.\n";
		     exit(1);
		  }
		  map < string, IDTYPE > dummy;
		  initialGraph = readHypergraph(igraph, &dummy);
	       }
	    }
	    else if(!xmlStrcmp(cur->name, (const xmlChar *)"Rule"))	// transformation rule?
	    {
	       RULE *r = readRule(cur, interpretNAC);
	       if(r != NULL)
	       {
		  if(rules.count(r->getID()) > 0)
		  {
		     this->lastUsedID++;
		     r->setID(this->lastUsedID);
		  }
		  else if(r->getID() > this->lastUsedID)
		     this->lastUsedID = r->getID();
		  rules[r->getID()] = *r;
	       }
	    }
	    else
	    {
	       //cerr << "Ignoring XML node " << cur->name << "\n";
	    }
	    cur = cur->next;
	 }

	 if(NULL == initialGraph)	// no initial graph specified
	 {
	    cout << "Input file does not specify an initial graph.\n";
	    exit(1);
	 }
      
         if(debug) initialGraph->setLastUsedID(this->lastUsedID);
	 GraphGrammar.setStartgraph(initialGraph);
	 GraphGrammar.setRules(&rules);
      }

      HYPERGRAPH *GTXL_NAC_READER::readHypergraph(xmlNodePtr root,
						  map < string, IDTYPE > *xmlID2ID)
      {
	 // true if hypergraph=true is set
	 bool
	    isHypergraph(!xmlStrcmp
			 (xmlGetProp(root, (const xmlChar *)"hypergraph"),
			  (const xmlChar *)"true"));
	 const xmlChar *edgemodeName(xmlGetProp(root, (const xmlChar *)"edgemode"));

	 if(isHypergraph && !xmlStrcmp(edgemodeName, (const xmlChar *)"directed"))
	 {
	    cerr << "This program does not support directed hypergraphs.\n";
	    exit(1);
	 }

	 bool isDirected(false);
	 if(!xmlStrcmp(edgemodeName, (const xmlChar *)"directed"))
	    isDirected = true;
	 if(!xmlStrcmp(edgemodeName, (const xmlChar *)"defaultdirected"))
	    isDirected = true;

	 HYPERGRAPH *result = new HYPERGRAPH();
         if(!debug)
         {
	    this->lastUsedID++;
	    result->setID(this->lastUsedID);
         }

	 // start parsing the node/edge list:
	 xmlNodePtr cur = root->xmlChildrenNode;
	 while(NULL != cur)
	 {
	    // Node:
	    if(!xmlStrcmp(cur->name, (const xmlChar *)"node"))
	    {
	    }			// do nothing; unconnected nodes are not permitted in simple transformation rules,    and connected nodes are implied by IDs stored in edges
	    // (biconnected) edge
	    else if(!xmlStrcmp(cur->name, (const xmlChar *)"edge"))
	    {
	       xmlChar *fromXMLID = xmlGetProp(cur, (const xmlChar *)"from");
	       xmlChar *toXMLID = xmlGetProp(cur, (const xmlChar *)"to");

	       IDTYPE fromID, toID;
	       if(1 == xmlID2ID->count(string((char *)fromXMLID)))
	       {
		  fromID = (*xmlID2ID)[string((char *)fromXMLID)];
		  if(!result->hasVertex(fromID))
		  {
		     VERTEX tv;
		     tv.setID(fromID);
		     tv.setIdLabel(string((char *)fromXMLID));
		     result->addVertex(tv, fromID);
		  }
		  else if(fromID > this->lastUsedID)
		     this->lastUsedID = fromID;
	       }
	       else
	       {
		  this->lastUsedID++;
		  fromID = this->lastUsedID;
		  VERTEX tv;
		  tv.setID(fromID);
		  tv.setIdLabel(string((char *)fromXMLID));
		  result->addVertex(tv, fromID);
		  (*xmlID2ID)[string((char *)fromXMLID)] = fromID;
	       }
	       if(1 == xmlID2ID->count(string((char *)toXMLID)))
	       {
		  toID = (*xmlID2ID)[string((char *)toXMLID)];
		  if(!result->hasVertex(toID))
		  {
		     VERTEX tv;
		     tv.setID(toID);
		     tv.setIdLabel(string((char *)toXMLID));
		     result->addVertex(tv, toID);
		  }
		  else if(toID > this->lastUsedID)
		     this->lastUsedID = toID;
	       }
	       else
	       {
		  this->lastUsedID++;
		  toID = this->lastUsedID;
		  VERTEX tv;
		  tv.setID(toID);
		  tv.setIdLabel(string((char *)toXMLID));
		  result->addVertex(tv, toID);
		  (*xmlID2ID)[string((char *)toXMLID)] = toID;
	       }

	       // find out if isdirected attribute is specified. If not, use default. Undirected Edges are represented    by two directed edges.
	       bool edgeIsDirected(isDirected);
	       if(!xmlStrcmp
		  (xmlGetProp(cur, (const xmlChar *)"isdirected"),
		   (const xmlChar *)"true"))
		  edgeIsDirected = true;
	       if(!xmlStrcmp
		  (xmlGetProp(cur, (const xmlChar *)"isdirected"),
		   (const xmlChar *)"false"))
		  edgeIsDirected = false;

	       string label(getLabel(cur));

	       if("" == label)	// warn if unlabeled edge was found
	       {
		  if(!quiet)
		     cerr << "WARNING: unlabeled edge found, using empty label.\n";
	       }
	       EDGE te;
	       this->lastUsedID++;
	       IDTYPE teID = this->lastUsedID;
	       te.setID(teID);
	       te.setLabel(label);
	       xmlChar *edgeXMLID = xmlGetProp(cur, (const xmlChar *)"id");
	       if(NULL != edgeXMLID) te.setIdLabel(string((char *)edgeXMLID));
	       list < IDTYPE > connectionList;
	       connectionList.push_back(fromID);
	       connectionList.push_back(toID);
	       result->addEdge(te, teID, connectionList);
	       if(NULL != edgeXMLID)
		  (*xmlID2ID)[string((char *)edgeXMLID)] = teID;
	       if(!edgeIsDirected)
	       {
		  te = EDGE();
		  this->lastUsedID++;
		  teID = this->lastUsedID;
		  te.setID(teID);
		  te.setLabel(label);
		  if(NULL != edgeXMLID) te.setIdLabel(string((char *)edgeXMLID));
		  connectionList.clear();
		  connectionList.push_back(toID);
		  connectionList.push_back(fromID);
		  result->addEdge(te, teID, connectionList);
		  xmlChar *edgeXMLID = xmlGetProp(cur, (const xmlChar *)"id");
		  if(NULL != edgeXMLID)
		     (*xmlID2ID)[string((char *)edgeXMLID)] = teID;
	       }
	    }
	    // Hyperedge
	    else if(!xmlStrcmp(cur->name, (const xmlChar *)"rel"))
	    {
	       if(!isHypergraph)	// only hypergraphs have rel edges
	       {
		  cout << "Hyperedge found in a non-hypergraph\n";
		  exit(1);
	       }
	       if(!xmlStrcmp
		  (xmlGetProp(cur, (const xmlChar *)"isdirected"),
		   (const xmlChar *)"true"))
	       {
		  // directed hyperedges are unsupported
		  cout <<
		     "Directed Hyperedge found. Directed Hyperedges are not supported.\n";
		  exit(1);
	       }
	       list < IDTYPE > connectionList;
	       xmlNodePtr relend = cur->xmlChildrenNode;
	       while(NULL != relend)
	       {
		  if(!xmlStrcmp(relend->name, (const xmlChar *)"relend"))
		  {
		     xmlChar *targetXMLID =
			xmlGetProp(relend, (const xmlChar *)"target");
		     IDTYPE targetID;
		     if(1 == xmlID2ID->count(string((char *)targetXMLID)))
		     {
			targetID = (*xmlID2ID)[string((char *)targetXMLID)];
			if(!result->hasVertex(targetID))
			{
			   VERTEX tv;
			   tv.setID(targetID);
			   tv.setIdLabel(string((char *)targetXMLID));
			   result->addVertex(tv, targetID);
			}
			else if(targetID > this->lastUsedID)
			   this->lastUsedID = targetID;
		     }
		     else
		     {
			this->lastUsedID++;
			targetID = this->lastUsedID;
			VERTEX tv;
			tv.setID(targetID);
			tv.setIdLabel(string((char *)targetXMLID));
			result->addVertex(tv, targetID);
			(*xmlID2ID)[string((char *)targetXMLID)] = targetID;
		     }
		     connectionList.push_back(targetID);
		  }
		  relend = relend->next;
	       }

	       string label = getLabel(cur);
	       if("" == label)	// warn if unlabeled edge was found
	       {
		  cerr << "WARNING: unlabeled hyperedge found, using empty label.\n";
	       }

	       this->lastUsedID++;
	       IDTYPE newEdgeID = this->lastUsedID;
	       EDGE te;
	       te.setLabel(label);
	       te.setID(newEdgeID);
	       xmlChar* edgeXMLID = xmlGetProp(cur, (const xmlChar *)"id");
	       if(NULL != edgeXMLID)
	       {
	          te.setIdLabel(string((char *)edgeXMLID));
		  (*xmlID2ID)[string((char *)edgeXMLID)] = newEdgeID;
	       }
	       result->addEdge(te, newEdgeID, connectionList);
	    }
	    cur = cur->next;
	 }
	 return result;
      }

      RULE* GTXL_NAC_READER::readRule(xmlNodePtr node, bool interpretNAC)
      {
	 // first, find the "mapping" child element:
	 xmlNodePtr cur = node->xmlChildrenNode;
	 while(NULL != cur && xmlStrcmp(cur->name, (const xmlChar *)"Mapping"))
	    cur = cur->next;

	 if(NULL == cur)	// no mapping specified?
	 {
	    cerr << "Transformation rule without mapping found\n";
	    exit(1);
	 }
	 // for each map element, register one ID for later use in hypergraph
	 // construction
	 map < string, IDTYPE > xmlID2ID;
	 // interface
	 map<IDTYPE, IDTYPE> interface;
	 
	 for(xmlNodePtr mapEl = cur->xmlChildrenNode; NULL != mapEl; mapEl = mapEl->next)
	 {
	    // ignore everything except MapElem nodes
	    if(!xmlStrcmp(mapEl->name, (const xmlChar *)"MapElem"))
	    {
	       string fromXmlID((char *)xmlGetProp(mapEl, (const xmlChar *)"from"));
	       string toXmlID((char *)xmlGetProp(mapEl, (const xmlChar *)"to"));
	       this->lastUsedID++;
	       IDTYPE id = this->lastUsedID;
	       xmlID2ID[fromXmlID] = id;
	       xmlID2ID[toXmlID] = id;
	       
	        // interface
		if(interface.find(id)==interface.end())
		     interface[id]=id;
	    }
	 }
	 // map reading complete.
	 // now read left hand side and right hand side
	 HYPERGRAPH *lhs = NULL;
	 HYPERGRAPH *rhs = NULL;
	 HYPERGRAPH *singleEdgeGraph = NULL;
	 map < IDTYPE, HYPERGRAPH > nac;
	 set < unsigned >lev;
	 for(cur = node->xmlChildrenNode; NULL != cur; cur = cur->next)
	 {
	    if(!xmlStrcmp(cur->name, (const xmlChar *)"LHS"))
	    {
	       lhs = readRuleGraph(cur, &xmlID2ID);
	       xmlNodePtr rulegraph = cur->xmlChildrenNode;
	       while(rulegraph != NULL
		     && xmlStrcmp(rulegraph->name, (const xmlChar *)"RuleGraph"))
		  rulegraph = rulegraph->next;
	       if(rulegraph != NULL)
	       {
		  for(xmlNodePtr curGC = rulegraph->xmlChildrenNode; curGC != NULL;
		      curGC = curGC->next)
		  {
		     if(interpretNAC && !xmlStrcmp(curGC->name, (const xmlChar *)"GraphCondition"))
		     {
			//XT cout << "NAC found\n";
			xmlNodePtr curGCelem, segMapElem;

			// construct mapping
			//
			map < string, IDTYPE > segID2ID;
			for(curGCelem = curGC->xmlChildrenNode;
			    curGCelem != NULL; curGCelem = curGCelem->next)
			{
			   if(!xmlStrcmp(curGCelem->name, (const xmlChar *)"Mapping"))
			   {
			      for(segMapElem = curGCelem->xmlChildrenNode;
				  segMapElem != NULL; segMapElem = segMapElem->next)
			      {
				 if(!xmlStrcmp
				    (segMapElem->name, (const xmlChar *)"MapElem"))
				 {
				    string segFromXmlID((char *)
							xmlGetProp(segMapElem,
								   (const xmlChar *)
								   "from"));
				    string segToXmlID((char *)
						      xmlGetProp(segMapElem,
								 (const xmlChar *)"to"));

				    IDTYPE id;
				    map < string, IDTYPE >::iterator mapit;
				    mapit = xmlID2ID.find(segFromXmlID);
				    if(mapit != xmlID2ID.end())
				    {
				       id = mapit->second;
				       if(id > lastUsedID) this->lastUsedID = id;
				    }
				    else
				    {
				       mapit = xmlID2ID.find(segFromXmlID);
				       if(mapit != xmlID2ID.end())
				       {
					  id = mapit->second;
					  if(id > lastUsedID) this->lastUsedID = id;
				       }
				       else
				       {
				          this->lastUsedID++;
					  id = this->lastUsedID;
				       }
				    }
				    segID2ID[segFromXmlID] = id;
				    segID2ID[segToXmlID] = id;
				    
				    // interface
				    if(interface.find(id)==interface.end())
				     interface[id]=id;

				    //XT cout << "mapping from " << segFromXmlID;
				    //XT cout << " to " << segToXmlID << " (id " << id << ")\n";
				 }
			      }
			   }
			}

			// analyze rulegraph
			//
			singleEdgeGraph = readRuleGraph(curGC, &segID2ID);
			this->lastUsedID++;
			singleEdgeGraph->setID(this->lastUsedID);
			if(singleEdgeGraph != NULL)
			{
			   nac.insert(pair < IDTYPE,
				      HYPERGRAPH > (singleEdgeGraph->getID(),
						    *singleEdgeGraph));
			}

			//XT cout << "\n";
		     }
		  }
	       }
	    }
	    if(!xmlStrcmp(cur->name, (const xmlChar *)"RHS"))
	       rhs = readRuleGraph(cur, &xmlID2ID);

	    // read levels
	    //
	    if(!xmlStrcmp(cur->name, (const xmlChar *)"attr"))
	    {
	       string attr_type((char *)xmlGetProp(cur, (const xmlChar *)"name"));
	       if(attr_type == "level")
	       {
		  xmlNodePtr int_level = cur->xmlChildrenNode->xmlChildrenNode;
		  if(int_level != NULL)
		  {
		     //XT cout << (char*)(int_level->content) << "\n";
		     if(atoi((char *)(int_level->content)) >= 0)
			lev.insert(atoi((char *)(int_level->content)));
		  }
	       }
	    }
	 }

	 if(NULL == lhs)
	 {
	    cerr << "Transformation rule without left hand side found\n";
	    exit(1);
	 }
	 if(NULL == rhs)
	 {
	    cerr << "Transformation rule without right hand side found\n";
	    exit(1);
	 }
	 // get xml-ID of the rule:
	 string strRuleID;
	 xmlChar *ruleid = xmlGetProp(node, (const xmlChar *)"id");
	 if(NULL != ruleid)
	 {
	    strRuleID = string((char *)ruleid);
	 }
	 
	 RULE* r = new RULE();
	 r->setLeft(lhs);
	 r->setRight(rhs);
	 r->setName(strRuleID);
	 MATCH match_interface;
	 match_interface.setMatch(interface);
	 r->setInterface(match_interface);
	 this->lastUsedID++;
	 r->setID(this->lastUsedID);
	 if(interpretNAC)
	 {
	    r->setNAC(nac);
	 }
	 return r;
      }

      HYPERGRAPH *GTXL_NAC_READER::readRuleGraph(xmlNodePtr node,
						 map < string, IDTYPE > *xmlID2ID)
      {
         // a rule graph contains a (hyper-)graph and some information which is
         // ignored here.
         xmlNodePtr cur = node->xmlChildrenNode;
         while(NULL != cur && xmlStrcmp(cur->name, (const xmlChar *)"RuleGraph"))
	    cur = cur->next;
         if(NULL == cur)		// no RuleGraph in the LHS/RHS?
         {
	    cout << node->name << " element without RuleGraph element found.\n";
	    exit(1);
         }
   
         cur = cur->xmlChildrenNode;
   
         while(NULL != cur && xmlStrcmp(cur->name, (const xmlChar *)"Graph"))
	    cur = cur->next;

         if(NULL == cur)		// no graph in the RuleGraph?
         {
	    cout << "RuleGraph element without a graph found.\n";
	    exit(1);
         }

         return readHypergraph(cur, xmlID2ID);
      }

      string GTXL_NAC_READER::getLabel(xmlNodePtr node)
      {
         for(xmlNodePtr cur = node->xmlChildrenNode; NULL != cur; cur = cur->next)
         {
   	    if(!xmlStrcmp(cur->name, (const xmlChar *)"attr"))
	       if(!xmlStrcmp
	          (xmlGetProp(cur, (const xmlChar *)"name"), (const xmlChar *)"label"))
	       {
	          cur = cur->xmlChildrenNode;
	          while(NULL != cur && xmlStrcmp(cur->name, (const xmlChar *)"string"))
		     cur = cur->next;
	          if(NULL == cur)
		     return "";
	          else
		     return string((char *)cur->xmlChildrenNode->content);
	       }
         }
         return "";
      }
   };
};
using namespace algorithms;
using namespace algorithms::readers;

#endif
