#include "types.h"

#include <sstream>

Node::~Node()
{
}

ostream & operator<<(ostream &ostream, const Node *node)
{
  ostream << node->to_string ();
  return ostream;
}

Operator::Operator (const char *name, Node *op1, Node *op2) :
  name (name)
{
  this->operands.push_back (op1);
  if (op2)
    this->operands.push_back (op2);
}

Operator::Operator (const char *name, const std::vector<Node*> &operands) :
  name (name),
  operands (operands)
{
}

Operator::Operator (const char *name, const list<Node*> &operands) :
  name (name),
  operands (operands.begin(), operands.end())
{
}

std::string Operator::to_string() const
{
  std::ostringstream os;
  if (operands.size() == 1)
    os << name << "( " << operands[0] << " )";
  else if (operands.size() == 2)
    os << "( " << operands[0] << " " << name << " " << operands[1] << " )";
  else if (operands.size() > 2) {
    std::vector<Node*>::const_iterator it=operands.begin();
    os << "( " << (*it);
    ++it;
    for (; it != operands.end(); ++it)
    {
      os << " " << name << " " << (*it);
    }
    os << " )";
  }
  return os.str();
}

Operator* Operator::copy() const
{
  std::vector<Node*> ops;
  for (std::vector<Node*>::const_iterator it = operands.begin();
       it != operands.end();
       ++it) {
    Node *tmp = (*it)->copy();
    ops.push_back(tmp);
  }
  return new Operator (name.c_str(), ops);
}

Quantor::Quantor (const char *name, Identifier *id, Node *formula) :
  name (name),
  variable (id),
  formula (formula)
{
}

std::string Quantor::to_string() const
{
  std::ostringstream os;
  os << name  << "." << variable << "(" << formula << ")";
  return os.str();
}

Quantor* Quantor::copy() const
{
  return new Quantor (name.c_str(), variable->copy(), formula->copy());
}

Function::Function (const char *name, Identifier *id) :
  name (name),
  variable (id)
{
}

Label::Label (Identifier *id) :
  Function ("label", id)
{
}

std::string Label::to_string () const
{
  std::ostringstream os;
  os << name << "(" << variable << ")";
  return os.str();
}

Label* Label::copy() const
{
  return new Label (variable->copy());
}

Connection::Connection (const unsigned int number, Identifier *id) :
  Function ("connection", id),
  number (number)
{
}

std::string Connection::to_string () const
{
  std::ostringstream os;
  os << name << "[" << number << "]" << "(" << variable << ")";
  return os.str();
}

Connection* Connection::copy() const
{
  return new Connection (number, variable->copy());
}

Identifier::Identifier (const char *name) :
  name (name)
{
}

std::string Identifier::to_string() const
{
  std::ostringstream os;
  os << name;
  return os.str();
}

Identifier* Identifier::copy() const
{
  return new Identifier (name.c_str());
}

Variable::Variable(const IDTYPE edge) :
  edge (edge)
{
}

std::string Variable::to_string() const
{
  std::ostringstream os;
  os << "#_" << edge;
  return os.str();
}

Variable* Variable::copy() const
{
  return new Variable (edge);
}

Constant::Constant(const int value) :
  value (value)
{
}

std::string Constant::to_string() const
{
  std::ostringstream os;
  os << value;
  return os.str();
}

Constant* Constant::copy() const
{
  return new Constant (value);
}

Boolean::Boolean(const bool value) :
  value (value)
{
}

std::string Boolean::to_string() const
{
  std::ostringstream os;
  os << value;
  return os.str();
}

Boolean* Boolean::copy() const
{
  return new Boolean (value);
}
