/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "pnmlreader.h"

#include <libxml/xpathInternals.h>

#include <sstream>

namespace algorithms {

namespace readers {

PNMLREADER::PNMLREADER()
 : ALGORITHM()
{
}


PNMLREADER::~PNMLREADER()
{
  xmlXPathFreeContext (ctx);
  xmlFreeDoc (doc);
}

bool PNMLREADER::isPNML(char *filename)
{
  xmlDocPtr doc = xmlParseFile (filename);
  if (!doc) {
    return false;
  }

  if (xmlStrcmp(xmlDocGetRootElement(doc)->name, (const xmlChar *) "pnml")) {
    xmlFreeDoc (doc);
    return false;
  }

  return true;
}

void PNMLREADER::initialize (char *filename)
{
  doc = xmlParseFile (filename);

  if (!doc) {
    cerr << "Document not parsed successfully." << endl;
    throw runtime_error ("XML parse error");
  }

  if (xmlStrcmp(xmlDocGetRootElement(doc)->name, (const xmlChar *) "pnml")) {
    cerr << "Document of the wrong type, root node != pnml" << endl;
    xmlFreeDoc (doc);
    throw runtime_error ("Wrong document type");
  }

  ctx = xmlXPathNewContext (doc);
  xmlXPathRegisterNs(ctx, (const xmlChar *)"pnml", (const xmlChar *)"http://www.informatik.hu-berlin.de/top/pnml");
}

void PNMLREADER::createPlaces (PTNET &PtNet)
{
  xmlXPathObjectPtr xpathObj; 

  xpathObj = xmlXPathEvalExpression((const xmlChar *)"//pnml:place", ctx);

  if (!xpathObj || !xpathObj->nodesetval) {
    cerr << "no places in doc ..." << endl;
    xmlFreeDoc (doc);
    xmlXPathFreeContext (ctx);
    throw runtime_error ("error in document");
  }

  xmlNodeSetPtr nodes = xpathObj->nodesetval;
  xmlNodePtr cur;
  num_places = nodes->nodeNr;
  for (int i = 0; i < num_places; ++i) {
    //cerr << "adding place ";
    cur = nodes->nodeTab[i];
    string s = (char *)xmlGetProp(cur, (const xmlChar *)"id");
    pnmlToIDMapping[s] = i;
    //cerr << s << endl;
    PtNet.addPlace (i);

    ostringstream path;
    path << "//pnml:place[attribute::id=\"" << s << "\"]/pnml:initialMarking/pnml:text/text()";
    xmlXPathObjectPtr marking = xmlXPathEvalExpression((const xmlChar *)path.str().c_str(), ctx);
    if (marking && marking->nodesetval && marking->nodesetval->nodeNr) {
      istringstream value ((char *)marking->nodesetval->nodeTab[0]->content);
      int tokens = 0;
      value >> tokens;
      TOKEN t;
      t.setCount (tokens);
      vector<TOKEN> vt;
      vt.push_back (t);
      initialMarking.setMarking (i, vt);
    }
  }

  PtNet.setInitialMarking (initialMarking);

  xmlXPathFreeObject(xpathObj);
}

void PNMLREADER::createTransitions (PTNET &PtNet)
{
  xmlXPathObjectPtr xpathObj;

  xpathObj = xmlXPathEvalExpression((const xmlChar *)"//pnml:transition", ctx);

  if (!xpathObj || !xpathObj->nodesetval) {
    cerr << "no transitions in doc ..." << endl;
    xmlFreeDoc (doc);
    xmlXPathFreeContext (ctx);
    throw runtime_error ("error in document");
  }

  xmlNodeSetPtr nodes = xpathObj->nodesetval;
  xmlNodePtr cur;
  for (int i = 0; i < nodes->nodeNr; ++i) {
    //cerr << "adding transition ";
    cur = nodes->nodeTab[i];
    string s = (char *)xmlGetProp(cur, (const xmlChar *)"id");
    pnmlToIDMapping[s] = i + num_places;
    //cerr << s << endl;
    PtNet.addTransition (i + num_places, TRANSITION ());
  }

  xmlXPathFreeObject(xpathObj);

  xpathObj = xmlXPathEvalExpression((const xmlChar *)"//pnml:arc", ctx);

  if (!xpathObj || !xpathObj->nodesetval) {
    cerr << "no arcs in doc ..." << endl;
    xmlFreeDoc (doc);
    xmlXPathFreeContext (ctx);
    throw runtime_error ("error in document");
  }

  nodes = xpathObj->nodesetval;
  for (int i = 0; i < nodes->nodeNr; ++i) {
    //cerr << "adding arc ";
    cur = nodes->nodeTab[i];
    string source = (char *)xmlGetProp(cur, (const xmlChar *)"source");
    string target = (char *)xmlGetProp(cur, (const xmlChar *)"target");
    //cerr << source << " -> " << target << endl;
    PtNet.addTConnection (pnmlToIDMapping[source], pnmlToIDMapping[target]);
  }

  xmlXPathFreeObject(xpathObj);
}

void PNMLREADER::run (PTNET &PtNet)
{
  createPlaces (PtNet);
  createTransitions (PtNet);
}

};

};
