/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "reachability_analyse.h"
#include "gxlreader.h"
#include "general_marking_reader.h"
#include "coverability.h"
#include "pnmlreader.h"

#include "logic_lex_yacc/types.h"

#include <fstream>
#include <typeinfo>
using namespace std;

namespace scenarios {

REACHABILITY_ANALYSE::REACHABILITY_ANALYSE()
  : SCENARIO()
{
  this->name = "reachability_analyse";

  this->inspec.push_back(pair<string, DATA_FORMAT>("petrigraph", GXL_PETRIGRAPH));
  this->inspec.push_back(pair<string, DATA_FORMAT>("logic", PLAIN_TEXT));
  this->outspec.push_back(pair<string, DATA_FORMAT>("marking", PLAIN_TEXT));

  this->call_descriptor = "Usage: augur -sc=reachability_analyse [options] logic petrigraph\n\n\
logic\t\tformula\n\
petrigraph\tinput petrigraph\n\
marking\t";
  this->cov = 0;
}


REACHABILITY_ANALYSE::~REACHABILITY_ANALYSE()
{
}

// iterate over the tokens and collect the count
long REACHABILITY_ANALYSE::getMarking (PTNET net, IDTYPE edge)
{
  vector<TOKEN> tokens = net.getInitialMarking().getMarking(edge);
  long marking = 0;
  for (vector<TOKEN>::iterator current_token = tokens.begin();
       current_token != tokens.end();
       ++current_token) {
    marking += current_token->getCount();
  }
  return marking;
}

LPMODEL REACHABILITY_ANALYSE::makeL0 (PTNET pn, const BoundSet &bounds)
{
  vector<IDTYPE> variables;

  //cerr << "creating L0 with formula: " << formula << endl;
  // variables for every edge (M)
  int num_places = 0;
  set<IDTYPE> places = pn.getPlaces();
  for (set<IDTYPE>::const_iterator it = places.begin();
       it != places.end();
       ++it) {
    variables.push_back (*it);
    num_places++;
  }
  // variables for every transition (X)
  map<IDTYPE, TRANSITION> transitions = pn.getTransitions();
  for (map<IDTYPE, TRANSITION>::const_iterator it = transitions.begin();
       it != transitions.end();
       ++it) {
    variables.push_back (it->first);
  }
  
  LPMODEL l(variables, num_places);

  l.set_int (variables);

  // fill rows
  for (map<IDTYPE, TRANSITION>::const_iterator it = transitions.begin();
       it != transitions.end();
       ++it) {
    map<IDTYPE, double> factors;
    vector<TCONNECTION> postset = pn.getPostset (it->first);
    vector<TCONNECTION> preset = pn.getPreset (it->first);
    for (vector<TCONNECTION>::const_iterator si = preset.begin();
         si != preset.end();
         ++si) {
      factors[si->ID] = - si->weight;
    }
    for (vector<TCONNECTION>::const_iterator si = postset.begin();
        si != postset.end();
        ++si) {
      if (factors.count (si->ID) == 0)
        factors[si->ID] = 0;
      factors[si->ID] += si->weight;
      if (factors[si->ID] == 0)
        factors.erase (si->ID);
    }
    l.set_column (it->first, factors);
  }

  // initial marking
  for (set<IDTYPE>::const_iterator it = places.begin();
       it != places.end();
       ++it) {
    map<IDTYPE, double> factors;
    factors[*it] = -1;
    l.set_column (*it, factors);
    l.set_constraint (*it, LPMODEL::CONSTRAINT_EQ);
    long marking = getMarking (pn, *it);
    //cerr << "marking of " << *it << ": " << marking << endl;
    l.set_rhs (*it, - marking);
  }

  for (BoundSet::const_iterator current_bound = bounds.begin();
       current_bound != bounds.end();
       ++current_bound) {
    Bounds b = current_bound->second;
    double lower = (b.lower != INT_MIN) ? b.lower : -l.INFINITE;
    double upper = (b.upper != INT_MAX) ? b.upper : l.INFINITE;
    l.set_bounds (current_bound->first, lower, upper);
  }
  
  return l;
}

bool REACHABILITY_ANALYSE::setHasID (const vector<TCONNECTION> &set, const IDTYPE &id)
{
  for (vector<TCONNECTION>::const_iterator it = set.begin();
       it != set.end();
       ++it) {
    if (it->ID == id)
      return true;
  }
  return false;
}

LPMODEL REACHABILITY_ANALYSE::makeFLP(PTNET pn)
{
  vector<IDTYPE> variables;

  // variables for every edge (Y)
  set<IDTYPE> places = pn.getPlaces();
  for (set<IDTYPE>::const_iterator it = places.begin();
       it != places.end();
       ++it) {
    variables.push_back (*it);
  }
  map<IDTYPE, TRANSITION> transitions = pn.getTransitions();
  
  LPMODEL l(variables, 0);
//  LPMODEL l(variables, transitions.size() + 1); // + 1 for M0

  // fill rows
  for (map<IDTYPE, TRANSITION>::const_iterator it = transitions.begin();
       it != transitions.end();
       ++it) {
    map<IDTYPE, double> factors;
    vector<TCONNECTION> postset = pn.getPostset (it->first);
    vector<TCONNECTION> preset = pn.getPreset (it->first);
    //cerr << "transition " << it->first << ": preset = " << preset.size() << "; postset = " << postset.size() << endl;
    for (vector<TCONNECTION>::const_iterator si = preset.begin();
         si != preset.end();
         ++si) {
      int num = 0;
      if (setHasID (postset, si->ID))
        num = preset.size() - 1;
      else
        num = -1;
      if (num)
        factors[si->ID] = num;
    }
    for (vector<TCONNECTION>::const_iterator si = postset.begin();
        si != postset.end();
        ++si) {
      if (!setHasID (preset, si->ID))
        factors[si->ID] = preset.size();
    }
    l.add_row (factors, LPMODEL::CONSTRAINT_GE, 0);
  }

  // initial marking
  map<IDTYPE, double> factors;
  for (set<IDTYPE>::const_iterator it = places.begin();
       it != places.end();
       ++it) {
    long marking = getMarking (pn, *it);
    if (marking)
      factors[*it] = marking;
  }
  l.add_row (factors, LPMODEL::CONSTRAINT_GE, 0.1);

  // add an empty dummy row that gets changed by changeFLP later
  map<IDTYPE, double> empty;
  l.add_row (empty, LPMODEL::CONSTRAINT_LE, 0);
  return l;
}

void REACHABILITY_ANALYSE::changeFLP (LPMODEL &input, const map<IDTYPE, double> &marking, PTNET pn)
{
  map<IDTYPE, double> factors;
  set<IDTYPE> places = pn.getPlaces();
  for (set<IDTYPE>::const_iterator it = places.begin();
       it != places.end();
       ++it) {
    map<IDTYPE, double>::const_iterator mit = marking.find (*it);
    if (mit != marking.end() && mit->second != 0)
      factors[*it] = mit->second;
  }
  input.remove_last_row ();
  input.add_row (factors, LPMODEL::CONSTRAINT_EQ, 0);
}

void REACHABILITY_ANALYSE::printVariables (ostream &os, const map<IDTYPE, double> &variables)
{
  os << "variables: " << endl;
  for (map<IDTYPE, double>::const_iterator it = variables.begin();
       it != variables.end();
       ++it) {
    os << it->first << ": " << it->second << endl;
  }
}

bool REACHABILITY_ANALYSE::trapCheck (PTNET pn, const BoundSet &bounds)
{
  LPMODEL lp = makeL0 (pn, bounds);

  LPMODEL::RESULT res = lp.solve();
  LPMODEL fp = makeFLP (pn);
  
  while (res == LPMODEL::RESULT_OPTIMAL) {
    map<IDTYPE, double> result_var = lp.get_variables ();
    //printVariables (cerr, result_var);
    changeFLP (fp, result_var, pn);
    LPMODEL::RESULT fpres = fp.solve ();
    switch (fpres) {
      case LPMODEL::RESULT_INFEASIBLE:
        cout << "FLP infeasible, marking may be reachable!" << endl;
        return false;
        break;
      case LPMODEL::RESULT_ERROR:
        cerr << "Internal error solving FLP, giving up!" << endl;
        return false;
        break;
      case LPMODEL::RESULT_OPTIMAL:
        cout << "FLP solvable" << endl;
        map<IDTYPE,double> flp_result = fp.get_variables ();
        map<IDTYPE,double> factors;
        cout << "found trap: ";
        //printVariables (cerr, flp_result);
        for (map<IDTYPE, double>::const_iterator it = flp_result.begin();
             it != flp_result.end();
             ++it) {
          if (it->second > 0) {
            factors[it->first] = 1;
            cout << it->first << " ";
          }
        }
        cout << endl;
        lp.add_row (factors, LPMODEL::CONSTRAINT_GE, 1);
        res = lp.solve ();
        break;
    }
  }
  if (res != LPMODEL::RESULT_OPTIMAL) {
    return true;
  }
  return false;
}

bool REACHABILITY_ANALYSE::canUseCoverability (const BoundSet &bounds)
{
  for (BoundSet::const_iterator current_bound = bounds.begin();
       current_bound != bounds.end();
       ++current_bound) {
    if (current_bound->second.upper != INT_MAX)
      return false;
  }
  return true;
}

MARKING REACHABILITY_ANALYSE::convertToMarking (const BoundSet &bounds)
{
  MARKING m;
  for (BoundSet::const_iterator current_bound = bounds.begin();
       current_bound != bounds.end();
       ++current_bound) {
    if (current_bound->second.lower != INT_MIN) {
      TOKEN t;
      t.setCount (current_bound->second.lower);
      vector<TOKEN> vt;
      vt.push_back (t);
      m.setMarking (current_bound->first, vt);
    }
  }
  return m;
}

bool REACHABILITY_ANALYSE::isUnreachable (PTNET pn, const BoundSet &bounds)
{
  cout << "checking: " << bounds << endl;
  bool unreachable = false;
  // first try trapCheck as it's usually quite a bit faster than COVERABILITY
  if (!trapCheck (pn,bounds)) {
    // cerr << "trapcheck failed" << endl;
    if (canUseCoverability (bounds)) {
      if (!cov) {
        cov = new algorithms::ptnet::coverability::COVERABILITY();
        cov->initialize (&pn);
      }
      MARKING m = convertToMarking (bounds);
      long psize;
      cout << "checking with COVERABILITY" << endl;
      // cerr << m << endl;
      unreachable = !cov->run (m, &psize);
    }
  } else {
    unreachable = true;
  }
  // cerr << "marking " << (unreachable ? "un" : "") << "reachable" << endl;
  return unreachable;
}

void REACHABILITY_ANALYSE::removeNot (Node *formula)
{
  if (typeid (*formula) != typeid (Operator))
    return;

  Operator *oper = (Operator *) formula;
  if (oper->name == "~") {
    // pull not down one level
    Operator *subop = dynamic_cast<Operator *>(oper->operands[0]);
    if (subop && subop->name == ">=") {
      oper->name = "<";
      oper->operands = subop->operands;
      delete subop;
    } else if (subop && subop->name == ">") {
      oper->name = "<=";
      oper->operands = subop->operands;
      delete subop;
    } else if (subop && subop->name == "<=") {
      oper->name = ">";
      oper->operands = subop->operands;
      delete subop;
    } else if (subop && subop->name == "<") {
      oper->name = ">=";
      oper->operands = subop->operands;
      delete subop;
    } else if (subop && subop->name == "=") {
      oper->name = "|";
      oper->operands.clear();
      oper->operands.push_back (new Operator ("<", subop->operands[0], subop->operands[1]));
      oper->operands.push_back (new Operator (">", subop->operands[0], subop->operands[1]));
      delete subop;
    } else if (subop && subop->name == "&") {
      oper->name = "|";
      oper->operands.clear ();
      for (vector<Node *>::iterator it = subop->operands.begin();
           it != subop->operands.end();
           ++it) {
        Node *current = new Operator ("~", *it, NULL);
        removeNot (current);
        oper->operands.push_back (current);
      }
      delete subop;
    } else if (subop && subop->name == "|") {
      oper->name = "&";
      oper->operands.clear ();
      for (vector<Node *>::iterator it = subop->operands.begin();
           it != subop->operands.end();
           ++it) {
        Node *current = new Operator ("~", *it, NULL);
        removeNot (current);
        oper->operands.push_back (current);
      }
      delete subop;
    }
  } else {
    for (vector<Node *>::iterator it = oper->operands.begin();
         it != oper->operands.end();
         ++it) {
      removeNot (*it);
    }
  }
}

vector<vector<unsigned int> > REACHABILITY_ANALYSE::permutate (vector<unsigned int> bounds)
{
  vector<vector<unsigned int> > permutations;
  if (!bounds.empty ()) {
    unsigned int last = bounds.back ();
    bounds.pop_back ();
    vector<vector<unsigned int> > sub_permutations = permutate (bounds);
    for (unsigned int i = 0; i <= last; i++) {
      if (sub_permutations.empty ()) {
        vector<unsigned int> new_perm;
        new_perm.push_back (i);
        permutations.push_back (new_perm);
      } else {
        for (vector<vector<unsigned int> >::iterator it = sub_permutations.begin();
             it != sub_permutations.end();
             ++it) {
          vector<unsigned int> new_perm(*it);
          new_perm.push_back (i);
          permutations.push_back (new_perm);
        }
      }
    }
  }
  return permutations;
}

// formula must not contain any ~ anymore
Node* REACHABILITY_ANALYSE::convertToDNF (Node *formula)
{
  //cerr << "convertToDNF called with formula: " << formula << endl;
  Operator *oper = dynamic_cast<Operator *>(formula);
  if (!oper) {
    return formula->copy ();
  }
  if (oper->name == ">="
      || oper->name == ">"
      || oper->name == "<="
      || oper->name == "<"
      || oper->name == "=") {
    return formula->copy ();
  }
  if (oper->name == "|") {
    vector<Node *> dnfs;
    for (vector<Node *>::iterator it = oper->operands.begin();
         it != oper->operands.end();
         ++it) {
      dnfs.push_back (convertToDNF (*it));
    }
    Operator *retval = new Operator ("|", dnfs);
    //cerr << "formula is a disjunction, returning: " << retval << endl;
    return retval;
  } else if (oper->name == "&") {
    vector<Node *> dnfs;
    for (vector<Node *>::iterator it = oper->operands.begin();
         it != oper->operands.end();
         ++it) {
      dnfs.push_back (convertToDNF (*it));
    }
    // inner one:  pij (conjunctions of the dnfs
    // middle one: disjunctions
    // outer one:  conjunction
    vector<vector<Node *> > pij;
    vector<unsigned int> ki;
    //cerr << "setting up pij" << endl;
    for (vector<Node *>::iterator it = dnfs.begin();
         it != dnfs.end();
         ++it) {
      vector<Node *> conjunctions;
      Operator *op = (Operator *)*it;
      if (op->name == "|") {
        for (vector<Node *>::iterator dnfit = op->operands.begin();
             dnfit != op->operands.end();
             ++dnfit) {
          //err << "pushing conjunction: " << *dnfit << endl;
          conjunctions.push_back ((*dnfit)->copy ());
        }
      } else {
        //cerr << "pushing conjunction: " << op << endl;
        conjunctions.push_back (op->copy ());
      }
      pij.push_back (conjunctions);
      ki.push_back (conjunctions.size () - 1);
    }
    unsigned int m = dnfs.size ();
    /*
    cerr << "vector to permutate: ";
    for (vector<unsigned int>::iterator it = ki.begin();
         it != ki.end();
         ++it)
      cerr << *it << " ";
    cerr << endl;
    */
    vector<vector<unsigned int> > permutations = permutate (ki);
    vector<Node *> operands;
    for (vector<vector<unsigned int> >::iterator permit = permutations.begin();
         permit != permutations.end();
         ++permit) {
      vector<Node *> conjunction_operands;
      for (unsigned int k = 0; k < m; k++) {
        //cerr << "looking up p(" << k << "," << (*permit)[k] << ")" << endl;
        Node *current = pij[k][(*permit)[k]];
        if ((typeid (*current) == typeid (Operator))
            && ((Operator*)current)->name == "&") {
          Operator *current_op = (Operator *) current;
          for (vector<Node *>::iterator it = current_op->operands.begin();
               it != current_op->operands.end();
               ++it) {
            conjunction_operands.push_back (*it);
          }
        } else {
          conjunction_operands.push_back (current);
        }
      }
      operands.push_back (new Operator ("&", conjunction_operands));
    }
    Node *retval;
    if (operands.size () == 1) {
      // only conjunctions, no disjunction
      retval = operands[0];
    } else {
      retval = new Operator ("|", operands);
    }
    //cerr << "returning: " << retval << endl;
    return retval;
  }
  return NULL;
}

void REACHABILITY_ANALYSE::run()
{
  if(coloured)
  {
   cout << "Error: LP Solve can not work with attributed Petri nets" << endl;
   exit(1);
  }

  writePPID();

  // algorithms::readers::MARKINGSREADER mr;
  GENERAL_MARKING_READER mr;
  mr.initialize((char*)this->filemap["logic"].c_str());
  Node *formula = NULL;
  mr.run (&formula);

  PTNET pn;

  if (algorithms::readers::PNMLREADER::isPNML((char*)(this->filemap["petrigraph"]).c_str())) {
    cout << "input Petrinet in PNML format" << endl;
    algorithms::readers::PNMLREADER reader;
    reader.initialize((char*)(this->filemap["petrigraph"]).c_str());
    reader.run(pn);
  } else {
    cout << "input Petrinet not in PNML format, assuming GXL" << endl;
    algorithms::readers::GXLREADER reader;
    reader.initialize((char*)(this->filemap["petrigraph"]).c_str());
    PETRIGRAPH pg;
    GTS gts;
    reader.run(pg, gts);
    pn = (PTNET)pg;
  }

  // cerr << "PN:" << endl << "======" << endl << pn << endl;

  cout << "formula to check: " << formula << endl;
  //cerr << "removing ~ from formula" << endl;
  removeNot(formula);
  //cerr << "formula without ~: " << formula << endl;
  Operator *dnf = (Operator *)convertToDNF (formula);
  cout << "formula in DNF: " << dnf << endl;

  vector<BoundSet> conjunctions = simplify (dnf);

  vector<pair<BoundSet, bool> > results;
  for (vector<BoundSet>::iterator it = conjunctions.begin();
       it != conjunctions.end();
       ++it) {
    bool unreachable = isUnreachable (pn, *it);
    pair<BoundSet, bool> current;
    current.first = *it;
    current.second = unreachable;
    results.push_back (current);
  }
  bool unreachable = true;
  for (vector<pair<BoundSet, bool> >::iterator it = results.begin();
       it != results.end();
       ++it) {
    // cout << (it->second ? "un" : "") << "reachable: " << it->first << endl << endl;
    unreachable &= it->second;
  }
  if (unreachable)
    cerr << endl << "Marking(s) unreachable!" << endl;
  else
    cerr << endl << "Marking(s) might be reachable!" << endl;
}

};
