#include <iostream>
#include <sstream>

#include "lpmodel.h"

LPMODEL::LPMODEL (const vector<IDTYPE> &variables, int num_rows)
{
  lp = make_lp (num_rows, variables.size ());
  set_verbose (lp, SEVERE);
  int i = 1;
  for (vector<IDTYPE>::const_iterator it = variables.begin();
       it != variables.end();
       ++it) {
    ostringstream os;
    os << *it;
    set_col_name (lp, i, (char *)os.str().c_str());
    variable_mapping[*it] = i;
    reverse_mapping[i] = *it;
    ++i;
  }

  INFINITE = get_infinite (lp);
}

LPMODEL::LPMODEL (const LPMODEL &model) :
  variable_mapping (model.variable_mapping),
  reverse_mapping (model.reverse_mapping),
  INFINITE (model.INFINITE)
{
  // FIXME: this seems to throw away all constraint types and right hand sides
  // ergo: don't use it
  lp = copy_lp (model.lp);
}

LPMODEL::~LPMODEL ()
{
  delete_lp (lp);
}

bool LPMODEL::add_row (const map<IDTYPE, double> &factors, CONSTRAINT_TYPE type, double rhs)
{
  int col_count = factors.size ();
  double row[col_count];
  int colno[col_count];
  int i = 0;
  for (map<IDTYPE, double>::const_iterator it = factors.begin();
       it != factors.end();
       ++it) {
    row[i] = it->second;
    colno[i] = variable_mapping[it->first];
    // cerr << "adding column " << variable_mapping[it->first] << ": " << it->first << " -> " << it->second << endl;
    ++i;
  }
  return (add_constraintex (lp, col_count, row, colno, type, rhs) == TRUE);
}

bool LPMODEL::add_row (const map<IDTYPE, int> &factors, CONSTRAINT_TYPE type, double rhs)
{
  map<IDTYPE, double> real_factors;
  for (map<IDTYPE, int>::const_iterator it = factors.begin();
       it != factors.end();
       ++it) {
    real_factors[it->first] = (double)it->second;
  }
  return add_row (real_factors, type, rhs);
}

bool LPMODEL::set_constraint (const IDTYPE &row, CONSTRAINT_TYPE type)
{
  return (set_constr_type (lp, variable_mapping[row], type) == TRUE);
}

bool LPMODEL::set_rhs (const IDTYPE &row, double rhs)
{
  return (set_rh (lp, variable_mapping[row], rhs) == TRUE);
}

bool LPMODEL::set_column (const IDTYPE &transition, const map<IDTYPE, double> &factors)
{
  int row_count = factors.size ();
  double col[row_count];
  int rowno[row_count];
  int i = 0;
  for (map<IDTYPE, double>::const_iterator it = factors.begin();
       it != factors.end();
       ++it) {
    col[i] = it->second;
    rowno[i] = variable_mapping[it->first];
    // cerr << "adding row  " << it->first << ": " << transition << "-> " << it->second << endl;
    ++i;
  }
  return (set_columnex (lp, variable_mapping[transition], row_count, col, rowno) == TRUE);
}

void LPMODEL::set_int (const vector<IDTYPE> &variables)
{
  for (vector<IDTYPE>::const_iterator it = variables.begin();
       it != variables.end();
       ++it) {
    ::set_int (lp, variable_mapping[*it], TRUE);
  }
}

bool LPMODEL::set_bounds (const IDTYPE &variable, double lower, double upper)
{
  return (::set_bounds (lp, variable_mapping[variable], lower, upper) == TRUE);
}

LPMODEL::RESULT LPMODEL::solve ()
{
  /*
  cout << "solving model:" << endl;
  write_LP (lp, stderr);
  */
  int res = ::solve (lp);
  if ((res == RESULT_OPTIMAL) || ( res == RESULT_INFEASIBLE))
    return (RESULT)res;
  else {
    cerr << "unknown return value of lpmodel solve: " << res << endl;
    return RESULT_ERROR;
  }
}

map<IDTYPE, double> LPMODEL::get_variables ()
{
  double *variables = new double[variable_mapping.size()];
  ::get_variables (lp, variables);
  map<IDTYPE, double> retval;
  for (unsigned int i = 0; i < variable_mapping.size(); ++i) {
    retval[reverse_mapping[i + 1]] = variables[i];
  }
  delete[] variables;
  return retval;
}

void LPMODEL::remove_last_row ()
{
  del_constraint (lp, get_Nrows (lp));
}

