/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "simplify_formula.h"

#include <set>
#include <iostream>
#include <typeinfo>

Bounds convertSingleConstraint (Operator *formula, const Bounds &current)
{
  Variable *v = (Variable *)formula->operands[0];
  Constant *c = (Constant *)formula->operands[1];
  Bounds b (current);
  if (formula->name == "=") {
    if (b.lower != INT_MIN || b.upper != INT_MAX) {
      cerr << "invalid duplicate bounds specified for " << v->edge << endl;
    }
    b.lower = c->value;
    b.upper = c->value;
  } else if (formula->name == ">=") {
    if (b.lower != INT_MIN) {
      cerr << "duplicate bounds specified for " << v->edge << endl;
    }
    b.lower = c->value;
  } else if (formula->name == ">") {
    if (b.lower != INT_MIN) {
      cerr << "duplicate bounds specified for " << v->edge << endl;
    }
    b.lower = c->value + 1;
  } else if (formula->name == "<=") {
    if (b.upper != INT_MAX) {
      cerr << "duplicate bounds specified for " << v->edge << endl;
    }
    b.upper = c->value;
  } else if (formula->name == "<") {
    if (b.upper != INT_MAX) {
      cerr << "duplicate bounds specified for " << v->edge << endl;
    }
    b.upper = c->value - 1;
  } else {
    cerr << "wrong formulatype: " << formula << endl;
  }
  return b;
}

// formula has to be a conjunction over constraints or a single constraint
BoundSet convertToBounds (Operator *formula)
{
  BoundSet bounds;
  if (formula->name != "&") {
    // single constraint
    Variable *v = (Variable *)formula->operands[0];
    Bounds b;
    bounds[v->edge] = convertSingleConstraint (formula, b);
    return bounds;
  }
  for (vector<Node *>::iterator current_operand = formula->operands.begin();
       current_operand != formula->operands.end();
       ++current_operand) {
    Operator *op = (Operator *)(*current_operand);
    Variable *v = (Variable *)op->operands[0];
    Bounds b;
    if (bounds.count (v->edge) == 1) {
      b = bounds[v->edge];
    }
    bounds[v->edge] = convertSingleConstraint (op, b);
  }
  return bounds;
}

ostream& operator<<(ostream &ostream, const BoundSet &bounds)
{
  ostream << "bound set: ";
  for (BoundSet::const_iterator current = bounds.begin();
       current != bounds.end();
       ++current) {
    ostream << current->first << "[";
    if (current->second.lower == INT_MIN)
      ostream << "MIN; ";
    else
      ostream << current->second.lower << ";";
    if (current->second.upper == INT_MAX)
      ostream << "MAX], ";
    else
      ostream << current->second.upper << "], ";
  }
  return ostream;
}

bool isSuperset (const Bounds &super, const Bounds &sub)
{
  return ((super.lower <= sub.lower) && (super.upper >= sub.upper));
}

bool isAdjacent (const Bounds &bound1, const Bounds &bound2)
{
  Bounds left, right;
  if (bound1.lower < bound2.lower) {
    left = bound1;
    right = bound2;
  } else {
    left = bound2;
    right = bound1;
  }
  return ((right.lower - left.upper) <= 1);
}

// pre: isAdjacent (bound1, bound2) = true
Bounds mergeSingleBounds (const Bounds &bound1, const Bounds &bound2)
{
  if (bound1.lower < bound2.lower) {
    return Bounds (bound1.lower, bound2.upper);
  } else {
    return Bounds (bound2.lower, bound1.upper);
  }
}

vector<BoundSet> mergeBounds (const BoundSet &a, const BoundSet &b, bool *changed)
{
  cerr << "merge bounds called with sets:" << endl << a << endl << b << endl;
  cerr << endl;
  vector<BoundSet> retval;
  bool superset = false;
  bool subset = false;
  bool adjacent_merged = false;
  BoundSet reta;
  BoundSet retb;
  if (a.size () != b.size ()) {
    // check if the smaller of a/b is a true superset of the bigger one
    BoundSet smaller;
    BoundSet bigger;
    if (a.size () < b.size()) {
      smaller = a;
      bigger = b;
    } else {
      smaller = b;
      bigger = a;
    }
    for (BoundSet::const_iterator it = smaller.begin();
         it != smaller.end();
         ++it) {
      if (bigger.count (it->first) == 0) {
        cerr << "variables don't match, not merging" << endl;
        goto no_change;
      }
      Bounds ba = it->second;
      Bounds bb = bigger.find (it->first)->second;
      if ((ba == bb) || isSuperset (ba, bb)) {
        reta[it->first] = ba;
      } else {
        goto no_change;
      }
    }
    *changed = true;
    retval.push_back (reta);
    cerr << "merged to " << reta << endl;
    return retval;
  }
  for (BoundSet::const_iterator it = a.begin();
       it != a.end();
       ++it) {
    if (b.count (it->first) == 0) {
      cerr << "variables don't match, not merging" << endl;
      goto no_change;
    } else {
      Bounds ba = it->second;
      Bounds bb = b.find (it->first)->second;
      if (ba == bb) {
        reta[it->first] = ba;
        retb[it->first] = ba;
      } else if (isSuperset (ba, bb)) {
        if (subset || adjacent_merged)
          goto no_change;
        reta[it->first] = ba;
        retb[it->first] = ba;
        *changed = true;
        superset = true;
      } else if (isSuperset (bb, ba)) {
        if (superset || adjacent_merged)
          goto no_change;
        reta[it->first] = bb;
        retb[it->first] = bb;
        *changed = true;
        subset = true;
      } else if (isAdjacent (ba, bb)) {
        if (subset || superset || adjacent_merged)
          goto no_change;
        Bounds merged = mergeSingleBounds (ba, bb);
        reta[it->first] = merged;
        retb[it->first] = merged;
        *changed= true;
        adjacent_merged = true;
      }
    }
  }
  retval.push_back (reta);
  if (!changed)
    retval.push_back (retb);
  else {
    cerr << "merged to " << reta << endl;
  }
  return retval;

no_change:
  retval.push_back (a);
  retval.push_back (b);
  cerr << "nothing to merge" << endl;
  return retval;
}

void mergeOperators (Node *formula)
{
  if (typeid (*formula) != typeid (Operator))
    return;

  Operator *oper = (Operator *) formula;
  vector<Node *> mergedOperands;
  for (vector<Node *>::iterator it = oper->operands.begin();
       it != oper->operands.end();
       ++it) {
    mergeOperators (*it);
    Operator *child = dynamic_cast<Operator *>(*it);
    if (child && oper->name == child->name) {
      // merge
      for (vector<Node *>::iterator childit = child->operands.begin();
           childit != child->operands.end();
           ++childit) {
        mergedOperands.push_back (*childit);
      }
    } else {
      mergedOperands.push_back (*it);
    }
  }
  oper->operands = mergedOperands;
}

// formula has to be in DNF
vector<BoundSet> simplify (Operator *formula)
{
  mergeOperators (formula);
  vector<BoundSet> all_bounds;
  if (formula->name == "|") {
    for (vector<Node *>::iterator current = formula->operands.begin();
         current != formula->operands.end();
         ++current) {
      cerr << "converting formula: " << *current << " to bounds." << endl;
      BoundSet bounds = convertToBounds ((Operator *)(*current));
      all_bounds.push_back (bounds);
      cerr << bounds << endl;
    }
  } else {
    cerr << "converting formula: " << formula << " to bounds." << endl;
    BoundSet bounds = convertToBounds (formula);
    all_bounds.push_back (bounds);
    cerr << bounds << endl;
  }
  bool changed;
  do {
    changed = false;
    set<BoundSet> new_bounds;
    set<BoundSet> unmerged;
    cerr << "all_bounds size: " << all_bounds.size () << endl;
    if (all_bounds.size() == 1) {
      continue;
    }
    unmerged.insert (all_bounds.begin(), all_bounds.end());
    while (all_bounds.size () > 1) {
      BoundSet first = all_bounds.back ();
      all_bounds.pop_back ();
      for (vector<BoundSet>::iterator current = all_bounds.begin ();
           current != all_bounds.end();
           ++current) {
        bool this_changed = false;
        vector<BoundSet> merged = mergeBounds (first, *current, &this_changed);
        if (this_changed) {
          new_bounds.insert (merged.begin(), merged.end());
          unmerged.erase (first);
          unmerged.erase (*current);
          changed = true;
        }
      }
    }
    all_bounds.pop_back ();
    all_bounds.insert (all_bounds.begin(), new_bounds.begin(), new_bounds.end());
    all_bounds.insert (all_bounds.begin(), unmerged.begin(), unmerged.end());
  } while (changed);
  cerr << "simplified." << endl;
  for (vector<BoundSet>::iterator current = all_bounds.begin();
       current != all_bounds.end();
       ++current) {
    cerr << *current << endl;
  }
  return all_bounds;
}


