/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "refinement.h"

#include "sample_reader.h"
#include "sample_writer.h"
#include "attr_example_checker.h"

#include "example_checker_col.h"
#include "equivalence_constructor_col.h"

namespace scenarios {

REFINEMENT::REFINEMENT()
{
      this->ID = ALGORITHM_REFINEMENT;
      this->name = "refinement";

      this->inspec.push_back(pair<string, DATA_FORMAT>("grammar", GTXL));
      this->inspec.push_back(pair<string, DATA_FORMAT>("petrigraph", GXL_PETRIGRAPH));
      this->inspec.push_back(pair<string, DATA_FORMAT>("counter-example", PLAIN_TEXT));
      this->outspec.push_back(pair<string, DATA_FORMAT>("sample", XML_GENERIC));
      this->outspec.push_back(pair<string, DATA_FORMAT>("petrigraph2", GXL_PETRIGRAPH));

      this->call_descriptor = "Usage: augur -sc=refinement [options] grammar petrigraph counter-example sample petrigraph2\n\n\
grammar\tgraph grammar in GTXL format\n\
petrigraph2\tinput abstraction for the refinement, a Petri graph in GXL format\n\
counter-example\tcounter-example, in TEXT format\n\
sample\tfile with current state of sample\n\
petrigraph2\toutput destination for the refinement, a Petri graph in GXL format\n\n";
}


REFINEMENT::~REFINEMENT()
{
}

void REFINEMENT::run()
{
      GTS grammar;
      ALGORITHM* gtxl_reader = this->load("call_gtxl_reader");
      gtxl_reader->initialize((char*)(this->filemap["grammar"]).c_str());
      gtxl_reader->run(grammar);

      PETRIGRAPH pg_original;
      ALGORITHM* gxl_reader = this->load("call_gxl_reader");
      gxl_reader->initialize((char*)(this->filemap["petrigraph"]).c_str());
      gxl_reader->run(pg_original, grammar);

      COUNTER_EXAMPLE_READER cer;
      cer.initialize((char*)(this->filemap["counter-example"]).c_str());
      COUNTER_EXAMPLE ce = cer.run();

      ALGORITHM* writer = this->load("call_gxl_writer");
      writer->initialize((char*)(this->filemap["petrigraph2"]).c_str());

      // check the counter-example
      SAMPLE s;
      int trans_num;
      bool result;

      if(!coloured)
      {
       EXAMPLE_CHECKER ech; 
       ech.initialize(&pg_original, ce.trans_ids);
       s = ech.run(&trans_num, &result);
      }
      else
      {
       EXAMPLE_CHECKER_COL ech;
       ech.initialize(&pg_original, ce.trans_ids, ce.bindings);
       s = ech.run(&trans_num, &result);
      }

      if(result)
      {
	cout << "The counter-example is real \n";
	exit(0);
      }	
      else
      {
	cout << "The example is spurious \n" << "Transition number " << abs(trans_num) << ", id: ";
	cout << ce.trans_ids[abs(trans_num)-1] << ", name: " << ce.trans_strings[abs(trans_num)-1];
	cout << ",\n" << "could not be fired \n";

	if(trans_num>0)
	 cout << "Structural refinement is needed" << endl;
	else
	 cout << "Refinement of attributes is needed" << endl;
      }

      // structural refinement
      if(trans_num>0)
      {
	// calculate the equivalence classes
	if(!coloured)
	{
		EQUIVALENCE_CONSTRUCTOR ec;
		ec.initialize(&pg_original, ce.trans_ids, &s);
		ec.run();
	}
	else
	{
		EQUIVALENCE_CONSTRUCTOR_COL ec;
		ec.initialize(&pg_original, ce.trans_ids, &s);
		ec.run();
	}
	
	// read the sample storage
	SAMPLE_STORAGE ss;
	ifstream ifsample;
	ifsample.open((char*)(this->filemap["sample"]).c_str());
	
	if(ifsample.is_open())
	{
		ifsample.close();
		SAMPLE_READER sr;
		sr.initialize((char*)(this->filemap["sample"]).c_str());
		ss = sr.run();
	}
	
	ss.m_list_storage.push_back(s);
	
	// set sample storage into SECOND_MATCHHANDLER_REFINEMENT
	storage.setData("second_matchhandler_refinement", NULL);
	storage.setData("second_matchhandler_refinement", &ss);
	
	// calculate the refinement
	PETRIGRAPH pg;
	init_time();
	ABSTRACTION_REFINEMENT refiner;
	refiner.initialize(&grammar, writer);
	refiner.run(&pg);
	
	cout << "time used: " << get_time() << " seconds\n";
	
	if(debug) cout << "\n" << pg << "\n";
	
	/*
	* promulgate Petri graph size
	*/
	cout << "edges: " << pg.getEdges().size()
		<< ", vertices: " << pg.getVertices().size()
		<< ", transitions: " << pg.getTransitions().size() << "\n";
	
	
	writer->run(pg, "augur2.out");
	
	// write sample	
	SAMPLE_WRITER sw;
	sw.initialize((char*)(this->filemap["sample"]).c_str());
	sw.run(ss);
      }
      // refinement of attributes 
      else
      {
	if(!coloured)
	{
		cout << "Possible conflict between attributed and not attributed parts\n";
		exit(1);
	}

	ATTR_EXAMPLE_CHECKER ae_checker;
	ae_checker.initialize(&pg_original, ce);
	bool ea_res = ae_checker.run(true);

	if(ea_res)
	{
		cout << "The counter-example is real\n";
		exit(0);
	}
	else
	{
		bool res = ae_checker.run(false);
		if(!res)
		{
			cout << "ERROR: Counter-Example can not be reproduced (with attributes)\n";
			exit(1);
		}

		// number of refinement steps
		for(unsigned int i = 0; i < attr_ref_count; i++)
		{
			storage.getEngine()->refine();
			res = ae_checker.run(false);

			cout << "(" << i+1 << ") Attributes have been refined\n";

			if(!res)
			{
				cout << "Spurious counter-example is eliminated\n";
				writer->run(pg_original, "augur2.out");
				exit(0);
			}
		}

		cout << "Attributes in the spurious counter-example could not be refined\n";
		exit(0);		
	}
      }
}

// namespace

}

