/***************************************************************************
 *   Copyright (C) 2007 by Salil Joshi   *
 *   hellfeuer@gmail.com   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "commkleene_differential.h"

namespace semilin {

COMMKLEENE_DIFFERENTIAL::COMMKLEENE_DIFFERENTIAL()
{
}

CFEXPRESSION* COMMKLEENE_DIFFERENTIAL::run(CFEXPRESSION *exp)
{
	/*basic differentiation rules.
	Whenever a + or concatenation is required, use simplifyPlus and simplifyConcat instead of doing it directly.
	*/
	
	switch (exp->type)
	{
		case (OP_PLUS):
		{
			CFEXPRESSION* dx = run(exp->lchild);
			CFEXPRESSION* dy = run(exp->rchild);
			return simplifyPlus(dx, dy);
			break;
		}
		case OP_CONCAT:
		{
			CFEXPRESSION* dx = run(exp->lchild);
			CFEXPRESSION* dy = run(exp->rchild);
			CFEXPRESSION* xdy = simplifyConcat(new CFEXPRESSION(exp->lchild), dy);
			CFEXPRESSION* ydx = simplifyConcat(new CFEXPRESSION(exp->rchild), dx);
			return simplifyPlus(xdy, ydx);
			break;
		}
		case VARIABLE:
		{
			if (exp->str_val == wrt) {
				return new CFEXPRESSION(STRING, "~"); //~ = epsilon
			} else {
				return new CFEXPRESSION(STRING, "/0"); //EMPTY SET
			}
			break;
		}
		case STRING:
		{
			return new CFEXPRESSION(STRING, "/0");
			break;
		}
		case OP_POWER:
		{
			CFEXPRESSION* dx = run(exp->lchild);
			if (exp->num_val == 2) {
				return simplifyConcat(new CFEXPRESSION(exp->lchild), dx);
			} else {
				return simplifyConcat(new CFEXPRESSION(OP_POWER, new CFEXPRESSION(exp->lchild), exp->num_val-1), dx);
			}
			break;
		}
		case OP_STAR:
		{
			return simplifyConcat(new CFEXPRESSION(exp), run(exp->lchild));
			break;
		}
		case OP_EQUAL:
		{
			return new CFEXPRESSION(OP_EQUAL, run(exp->lchild), run (exp->rchild));
			break;
		}
	}
}

void COMMKLEENE_DIFFERENTIAL::initialize(string wrtVar)
{
	wrt = wrtVar;
}

/*
Given two expressions 'left' and 'right', creates an expression 'left+right', but returns this in a simplified form. More simplifications here would be useful
*/
CFEXPRESSION* COMMKLEENE_DIFFERENTIAL::simplifyPlus(CFEXPRESSION* left, CFEXPRESSION* right)
{
	if (left == right) {// this comparison is potentially expensive, as it traverses the tree.
		delete left;
		return right;
	}
	if ((left->type == STRING) && (left->str_val == "/0")) {
		delete left;
		return right;
	}
	if ((right->type == STRING) && (right->str_val == "/0")) {
		delete right;
		return left;
	}
	
	return new CFEXPRESSION(OP_PLUS, left, right);
}

/*
Given two expressions 'left' and 'right', concatenates them: 'leftright', but returns this in a simplified form. More simplifications here would be useful
*/
CFEXPRESSION* COMMKLEENE_DIFFERENTIAL::simplifyConcat(CFEXPRESSION* left, CFEXPRESSION* right)
{
	if ((left->type == STRING) && (right->type == STRING)) {
		if ((left->str_val == "/0") || (right->str_val == "~")) {
			delete right;
			return left;
		} else if ((right->str_val == "/0") || (left->str_val == "~")) {
			delete left;
			return right;
		} else {
			CFEXPRESSION* result = new CFEXPRESSION(STRING, left->str_val + right->str_val);
			delete left;
			delete right;
			return result;
		}
	} else if (left->type == STRING) {
		if (left->str_val == "~") {
			delete left;
			return right;
		} else if (left->str_val == "/0") {
			delete right;
			return left;
		} else {
			return new CFEXPRESSION(OP_CONCAT, left, right);
		}
	} else if (right->type == STRING) {
		if (right->str_val == "~") {
			delete right;
			return left;
		} else if (right->str_val == "/0") {
			delete left;
			return right;
		} else {
			return new CFEXPRESSION(OP_CONCAT, left, right);
		}
	} else {
 		return new CFEXPRESSION(OP_CONCAT, left, right);
	}
		
}

COMMKLEENE_DIFFERENTIAL::~COMMKLEENE_DIFFERENTIAL()
{
}

}
