/*****************************************************************************
 * gxlreader.cpp: GXL (Graph eXchange Language) reader (source)
 *****************************************************************************
 * Copyright (C) 2004 Universitaet Stuttgart
 * $Id: gxlreader.cc,v 1.1.2.2 2005/02/14 16:53:54 koziouvi Exp $
 *
 * Authors: Nicolas Relange <relangns@studi.informatik.uni-stuttgart.de>
 *          Ingo Walther
 *          Lars Heinemann
 *          Barbara K�ig <koenigba@fmi.uni-stuttgart.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************/

#ifndef GXLREADER_CC
#define GXLREADER_CC

#if HAVE_CONFIG_H
#  include "config.h"
#endif

#include "gxlreader.h"
#include "error.h"
#include "common.h"
#include "hypergraph.cc"

#include <assert.h>
#include <cstdio>
#include <stdexcept>
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <list>

using  namespace std;

template<typename T,typename L>
GXLReader<T,L>::GXLReader(char* filename)
  : graph(NULL)
{
  doc = xmlParseFile(filename);

  if (doc == NULL )
    {
      fprintf(stderr,"Document not parsed successfully. \n");
      xmlFreeDoc(doc);
      exit(1);
    }

  if (xmlStrcmp(xmlDocGetRootElement(doc)->name, (const xmlChar *) "gxl")) 
      {
	fprintf(stderr,"document of the wrong type, root node != GTS");
	xmlFreeDoc(doc);
	exit(1);
      }

  graph = new Hypergraph<T,L>();
}


template<typename T,typename L>
GXLReader<T,L>::~GXLReader()
{
  delete graph;
  xmlFreeDoc(doc);
}


template<typename T,typename L>
Hypergraph<T,L>*
GXLReader<T,L>::read()
{
  xmlNodePtr root = xmlDocGetRootElement(doc);
  Error::info("GXLReader::read. Value of root",(const char*)root->name,4);
  assert(xmlStrEqual(root->name,(const xmlChar*)"gxl"));
  xmlNodePtr cur = root->xmlChildrenNode->next;
  Error::info("GXLReader::read. Value of first child",(const char*)cur->name,4);
  assert(xmlStrEqual(cur->name,(const xmlChar*)"graph"));

  // properties of the graph
  // hypergraph property
  bool isHypergraph(xmlStrEqual(xmlGetProp(cur,(const xmlChar *)"hypergraph"),(const xmlChar *)"true"));
  assert(isHypergraph);
  // edgemode property
  const xmlChar *edgemodeName(xmlGetProp(cur,(const xmlChar *)"edgemode"));
  bool isDirected(false);
  if(xmlStrEqual(edgemodeName,(const xmlChar *)"directed")) 
    isDirected = true;
  if(xmlStrEqual(edgemodeName,(const xmlChar *)"defaultdirected"))
    isDirected = true;
  assert(!isDirected);

  for(cur = cur->xmlChildrenNode; cur != NULL; cur = cur->next)
    {
      // node
      if(xmlStrEqual(cur->name,(const xmlChar *)"node"))
	parseVertex(cur);
      
      // hyperedge (rel)
      else if(xmlStrEqual(cur->name,(const xmlChar *)"rel"))
	parseEdge(cur);
    }
  return graph;
}


template<typename T,typename L>
void
GXLReader<T,L>::parseVertex(xmlNodePtr node)
{
  Error::info("parseVertex. Value of root node",(const char*)node->name,5);
  assert(xmlStrEqual(node->name,(const xmlChar*)"node"));

  xmlNodePtr cur;
  for(cur = node->xmlChildrenNode; cur != NULL; cur = cur->next)
    if(xmlStrEqual(cur->name,(const xmlChar *)"attr") && xmlStrEqual(xmlGetProp(cur,(const xmlChar*)"name"),(const xmlChar *)"vertex")) // check if it is really a vertex
      graph->addVertex(xmlCharToT(xmlGetProp(node,(const xmlChar *)"id")));
  
  return;
}


template<typename T,typename L>
void
GXLReader<T,L>::parseEdge(xmlNodePtr node)
{
  Error::info("parseEdge. Value of root node",(const char*)node->name,4);
  assert(xmlStrEqual(node->name,(const xmlChar*)"rel"));

  T id = xmlCharToT(xmlGetProp(node,(const xmlChar *)"id"));
  Error::info("parseEdge. id",id,4);
  L label;
  T targetId;
  unsigned startorder;

  // initiation of the lists for the order of the vertices
  list<T> connectionList,orderList;
  typename list<T>::iterator i,j;

  xmlNodePtr relend;
  for(relend = node->xmlChildrenNode; relend != NULL; relend = relend->next)
    {
      // label
      xmlNodePtr childNode;
      if(xmlStrEqual(relend->name,(const xmlChar *)"attr") &&
	 xmlStrEqual(xmlGetProp(relend,(const xmlChar*)"name"),(const xmlChar*)"label"))
	{
	  childNode = relend->xmlChildrenNode->next;
	  Error::info("parseEdge (label). Value of node",(const char*)childNode->name,4);
	  assert(xmlStrEqual(childNode->name,(const xmlChar*)"string"));
	  Error::info("parseEdge (label). Content of node",(const char*)childNode->xmlChildrenNode->content,4);
	  label = childNode->xmlChildrenNode->content;
	  continue;
	}
      
      // vertex
      if(xmlStrEqual(relend->name,(const xmlChar *)"relend") &&
	 xmlStrEqual(xmlGetProp(relend,(const xmlChar *)"role"),(const xmlChar *)"vertex"))
	{
	  targetId = xmlCharToT(xmlGetProp(relend,(const xmlChar *)"target"));
	  startorder = (unsigned) xmlCharToT(xmlGetProp(relend,(const xmlChar*)"startorder"));

	  // could be more efficient
	  for(j = orderList.begin(), i = connectionList.begin();
	      j != orderList.end() && *j < startorder;
	      ++j, ++i);
	  
	  orderList.insert(j,startorder);
	  connectionList.insert(i,targetId);
	}
    }
  
  // check that the orderList is correct
  int k;
  for(j = orderList.begin(), k = 0;
      j != orderList.end();
      ++j, ++k)
    assert( *j == k );
  
  if(xmlStrEqual(label,(const xmlChar*)"")) // warn if unlabeled edge was found
    {
      cerr << "WARNING: unlabeled hyperedge found, using empty label.\n";
    }

  graph->addEdge(connectionList,label,id);
  
  return;
}


template<typename T,typename L>
T
GXLReader<T,L>::xmlCharToT(const xmlChar* ch)
{
  Error::info("xmlCharToT. ch",(const char*)ch,4);
  T t;
  int i;
  for(i = 0, t = 0; ch[i] != '\0'; i++)
    if(ch[i] >= 48 && ch[i] <= 57)
      t = t*10 + ch[i] - 48;
  Error::info("xmlCharToT. t",t,4);
  return t;
}


// instantions of the template
template class GXLReader<idType,labelType>;

#endif // GXLREADER_CC
