/*****************************************************************************
 * hypergraph.cpp: representation of a hypergraph (source)
 *****************************************************************************
 * Copyright (C) 2004 Universitaet Stuttgart
 * $Id: hypergraph.cc,v 1.1.2.2 2005/02/14 16:53:54 koziouvi Exp $
 *
 * Authors: Nicolas Relange <relangns@studi.informatik.uni-stuttgart.de>
 *          Ingo Walther
 *          Lars Heinemann
 *          Barbara Knig <koenigba@fmi.uni-stuttgart.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************/

#ifndef HYPERGRAPH_CC
#define HYPERGRAPH_CC

#if HAVE_CONFIG_H
#  include "config.h"
#endif

#include "hypergraph.h"
#include "error.h"
#include "common.h"

#include <libxml/parser.h>
#include <assert.h>

using namespace std;

template<typename T,typename L>
Hypergraph<T,L>::Hypergraph()
  : vertices(std::set<T>()),
    edges(std::set<T>()),
    connectedV(std::map<T,std::list<T> >()),
    connectedE(std::map<T,std::multiset<T> >()),
    labels(std::map<T,L>()),
    labelToEdge(std::multimap<L,T>()),
    lastId(0)
{
}


template<typename T,typename L>
Hypergraph<T,L>::Hypergraph(const Hypergraph<T,L>& h)
  : vertices(h.vertices),
    edges(h.edges),
    connectedV(h.connectedV),
    connectedE(h.connectedE),
    labels(h.labels),
    labelToEdge(h.labelToEdge),
    lastId(h.lastId)
{
}


template<typename T,typename L>
Hypergraph<T,L>::~Hypergraph()
{
}  


template<typename T,typename L>
T
Hypergraph<T,L>::addVertex()
{
  T id = createNewId();
  return addVertex(id);
}


template<typename T,typename L>
T
Hypergraph<T,L>::addVertex(T id)
{
  if(0==vertices.count(id))
    {
      Error::info("addVertex. id",id,4);
      vertices.insert(id);
      connectedE.insert(pair<T,multiset<T> >(id,multiset<T>()));
    }
  return id;
}


template<typename T,typename L>
T
Hypergraph<T,L>::addEdge(list<T> connection, L label)
{
  T id = createNewId();
  return addEdge(connection,label,id);
}


template<typename T,typename L>
T
Hypergraph<T,L>::addEdge(list<T> connection, L label, T id)
{
  if(0==edges.count(id))
    {
      Error::info("addEdge. label",(const char*)label,4);
      Error::info("addEdge. id", id, 4);
      edges.insert(id);
      connectedV.insert(pair<T,list<T> >(id,connection));
      labels.insert(pair<T,L>(id,label));
      labelToEdge.insert(pair<L,T>(label,id));
      
      for(typename list<T>::iterator i = connection.begin();
	  i != connection.end();
	  i++)
	{
	  assert(1==vertices.count(*i)); // connection must not contain other IDs than vertices of this hypergraph
	  // find multiset of edges connected to vertex *i
	  typename map<T,multiset<T> >::iterator j = connectedE.find(*i);
	  j->second.insert(id);
	}
    }

  return id;
}


template<typename T,typename L>
void
Hypergraph<T,L>::uniongraphs(const Hypergraph<T,L>& h)
{
  // add vertices of h to this:
  for(typename set<T>::iterator i = h.vertices.begin();
      i != h.vertices.end();
      i++)
    {
      // assert vertice set are disjunct
      assert(0 == vertices.count(*i) );
      vertices.insert(*i);
    }
  // add edges and labels of h to this:
  for(typename set<T>::const_iterator j = h.edges.begin();
      j != h.edges.end();
      j++)
    {
      // assert edge sets are disjunct
      assert (0 == edges.count(*j));
      edges.insert(*j);
      labels.insert(*(h.labels.find(*j)));
    }
  // add the reverse mapping of labels
  for(typename multimap<L,T>::const_iterator m = h.labelToEdge.begin();
      m != h.labelToEdge.end();
      ++m)
    labelToEdge.insert(*m);
  // add connectivity mapping
  for(typename map<T,list<T> >::const_iterator k = h.connectedV.begin();
      k != h.connectedV.end();
      k++)
    {
      connectedV.insert(*k);
    }
  for(typename map<T,multiset<T> >::const_iterator l = h.connectedE.begin();
      l != h.connectedE.end();
      l++)
    {
      connectedE.insert(*l);
    }
}


template<typename T,typename L>
L
Hypergraph<T,L>::getLabel(T edge) const
  {
    assert(1 == edges.count(edge));

    return (*(labels.find(edge))).second;
  }


template<typename T,typename L>
std::set<T>
Hypergraph<T,L>::getVertices() const
{
  return vertices;
}


template<typename T,typename L>
std::set<T>
Hypergraph<T,L>::getEdges() const
{
  return edges;
}


template<typename T,typename L>
std::set<T>
Hypergraph<T,L>::getEdges(L label) const
{
  std::set<T> res;

  // First method: with the reverse labelling function labelToEdge
  // labelToEdges needs being good implemented with a comparison function for the string
//   typename std::multimap<L,T>::const_iterator i;
//   std::pair<typename std::multimap<L,T>::const_iterator,typename std::multimap<L,T>::const_iterator> range = this->labelToEdge.equal_range(label);
//   for(i = range.first;
//       i != range.second;
//       ++i)
//     res.insert(i->second);

  // Second method: with the labelling function labels
  // Easier to implement, but not so efficient
  typename std::map<T,L>::const_iterator j;
  for(j = this->labels.begin();
      j != this->labels.end();
      ++j)
    if( xmlStrEqual(label,j->second))
      res.insert(j->first);
  return res;
}


template<typename T,typename L>
list<T>
Hypergraph<T,L>::getConnectedVertices(T edge) const
{
  assert(1 == edges.count(edge));
  
  return connectedV.find(edge)->second;
}


template<typename T,typename L>
multiset<T>
Hypergraph<T,L>::getConnectedEdges(T vertex) const
{
  assert (1 == vertices.count(vertex));
  
  return connectedE.find(vertex)->second;
}

  
template<typename T,typename L>
map<T,T>
Hypergraph<T,L>::findMatching(Hypergraph<T,L>& h, MatchHandler<T>& handler)
{
  // initialise datastructures
  stack<typename set<T>::iterator> mappings;
  
  stack<T> unmapped_edges;
  for(typename set<T>::iterator i = h.edges.begin();
      i != h.edges.end();
      i++)
    unmapped_edges.push(*i);
  map<T,T> phi_v;
  map<T,T> phi_e;
  
  
  map<T,T> result;
  if(findNextMatching(h,mappings,unmapped_edges,phi_v,phi_e,handler)) // Does a matching exist?
    {
      // return union of phi_v and phi_e
      for(typename map<T,T>::iterator v = phi_v.begin();
	  v != phi_v.end();
	  v++)
	result.insert(*v);
	for(typename map<T,T>::iterator e = phi_e.begin();
	    e != phi_e.end();
	    e++)
	  result.insert(*e);
	return result;
    }
  else // no matching found
    {
      // findNextMatching should have cleared phi and mappings
      assert(0 == phi_v.size());
      assert(0 == phi_e.size());
      assert(0 == mappings.size());
      assert(unmapped_edges.size() == h.edges.size());
      return result; // return an empty map
    }
}


// private

template<typename T,typename L>
bool
Hypergraph<T,L>::findNextMatching(Hypergraph &h,
				  stack<typename set<T>::iterator> &mappings,
				  stack<T> &unmapped_edges,
				  map<T,T> &phi_v,
				  map<T,T> &phi_e,
				  MatchHandler<T>& matchhandler)
{
  // a matching was found?
  if(0 == unmapped_edges.size())
    {
      map<T,T> thematching;
      for(typename map<T,T>::iterator v = phi_v.begin();
	  v != phi_v.end();
	  v++)
	thematching.insert(*v);
      for(typename map<T,T>::iterator e = phi_e.begin();
	  e != phi_e.end();
	  e++)
	thematching.insert(*e);	
      return matchhandler.useThisMatch(thematching);
    }
  else // the matching is not yet complete
    {
      // edges and vertices in *this are named with prefix g,
      // edges and vertices in h are named with prefix h
	// get the next edge to be mapped
      T hedge = unmapped_edges.top();
      unmapped_edges.pop();
      // iterator on mappings stack iterates through all edges of g
      // to find an appropriate mapping
      mappings.push(edges.begin());
      // label of hedge:
      L hedgelabel = h.getLabel(hedge);
      // all vertices connected to the edge to be mapped
      list<T> hcV = h.getConnectedVertices(hedge);
      // iterate mappings.top() throung all edges of g
      while(mappings.top() != edges.end())
	{
	  T gedge = *(mappings.top());
	  // label of gedge:
	  L gedgelabel = getLabel(gedge);
	  // all vertices connected to the mapping target
	  list<T> gcV = getConnectedVertices(gedge);
	  // the label and cardinality of matching edges must be the same
	  if((hedgelabel != gedgelabel) || (gcV.size() != hcV.size()))
	    {
	      (mappings.top())++; // try another edge
	      continue; 
	    }
	  // store all new vertice mappings to delete them later
	  list<T> newMappedVertices;
	  bool vertexMatchingFailed(false);
	  

	  typename list<T>::iterator hv = hcV.begin();
	  typename list<T>::iterator gv = gcV.begin();
	  
	  while((!vertexMatchingFailed) && hv != hcV.end())
	    {
	      
	      typename map<T,T>::iterator phi_v_it = phi_v.find(*hv);
	      if(phi_v.end() == phi_v_it) // vertex is not yet mapped
		{
		  newMappedVertices.push_back(*hv);
		  phi_v.insert(pair<T,T>(*hv,*gv));
		}
	      else //vertex is already mapped
		if((*phi_v_it).second != *gv) //mismatch, abort
		  vertexMatchingFailed = true;
	      gv++;
	      hv++;
	    }
	  // now all vertices are mapped, or vertexMatchingFailed==true
	  if(!vertexMatchingFailed)
	      {
		phi_e.insert(pair<T,T>(hedge,gedge));
		if(findNextMatching(h,mappings,unmapped_edges,phi_v,phi_e,matchhandler)) return true;
		phi_e.erase(hedge);
	      }
	  // undo all changes in phi_v
	    for(typename list<T>::iterator v = newMappedVertices.begin();
		v != newMappedVertices.end();
		v++)
	      phi_v.erase(*v);
	    (mappings.top())++; // try next edge mapping
	  
	}
      unmapped_edges.push(hedge);
      mappings.pop();
      
      return false; // nothing found
    }  
}


template<typename T,typename L>
T
Hypergraph<T,L>::createNewId()
{
  ++lastId;
  while(vertices.count(lastId) || edges.count(lastId))
    ++lastId;
  return lastId;
}


template<typename T,typename L>
bool
Hypergraph<T,L>::strEqu(char* a,char* b) const
{
  int i;
  for(i = 0; a[i] == b[i]; i++)
    {
      if(a[i] == '\0')
	return true;
    }
  return false;
}


template<typename T,typename L>
std::ostream&
operator<< (std::ostream& os, const Hypergraph<T,L>& hg)
{
  os << "vertices: ";
  for(typename set<T>::const_iterator i = hg.vertices.begin();
      i != hg.vertices.end();
      i++)
    os << (*i) << " ";
  os << "  edges: ";
  for(typename set<T>::const_iterator i = hg.edges.begin();
      i!= hg.edges.end();
      i++)
    {
      os << (*i) << " <" << hg.getLabel(*i) << ">(";
      typename map<T,list<T> >::const_iterator j = hg.connectedV.find(*i);
      for(typename list<T>::const_iterator k = j->second.begin();
	  k != j->second.end();
	  k++)
	os << (*k) << " ";
      os << ") ";
    }
    return os;
}



// instantiations of the template
template class Hypergraph<idType,labelType>;
template class MatchHandler<idType>;

#endif // HYPERGRAPH_CC
