/*****************************************************************************
 * hypergraph.h: representaiton of a hypergraph (header)
 *****************************************************************************
 * Copyright (C) 2004 Universitaet Stuttgart
 * $Id: hypergraph.h,v 1.1.2.1 2004/08/30 15:18:23 koziouvi Exp $
 *
 * Authors: Nicolas Relange <relangns@studi.informatik.uni-stuttgart.de>
 *          Ingo Walther
 *          Lars Heinemann
 *          Barbara Knig <koenigba@fmi.uni-stuttgart.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************/

#ifndef HYPERGRAPH_H
#define HYPERGRAPH_H

#include <iostream>
#include <list>
#include <map>
#include <set>
#include <stack>

// forward declarations (see below)
template<typename T>
class MatchHandler;

template<typename T,typename L>
class Hypergraph;

template<typename T,typename L>
std::ostream& operator<< (std::ostream& os, const Hypergraph<T,L>& hg);


//! Representation of a hypergraph
/*! Hypergraph<T,L> is a hypergraph
 *  - with Ids of type T
 *  - and labels of type L
 *
 *  Make sure:
 *  - operator< and operator++ are defined for T
 *  - there is a constructor von T with 0 as an argument
 *  - operator== and operator!= are defined for L
 */
template<typename T,typename L>
class Hypergraph
{
 public:
  //! Constructor
  Hypergraph();
  //! Copy constructor
  Hypergraph(const Hypergraph<T,L>& h);
  //! Destructor
  ~Hypergraph();

  //! Adds a new vertex to the hypergraph
  /*! \return Id of the created vertex
   */
  T addVertex();
  
  //! Adds a new vertex with a specified Id to the hypergraph
  /*! The given Id must be unique.
   *  If Id is already a vertex of the graph, nothing is done
   *  (Exception should be raised ??)
   *  \param id Id of the new vertex
   *  \return Id of the created vertex
   */
  T addVertex(T id);

  //! Adds a new hyperedge to the hypergraph
  /*! \param connection List of vertices 
   *  \param label Label for the new edge
   *  \return Id of the new edge
   */
  T addEdge(std::list<T> connection, L label);

  //! Adds a new hyperedge to the hypergraph with a specified Id
  /*! The given Id must be unique.
   *  If Id is already an hyperedge of the graph, nothing is done
   *  (Exception should be raised ??)
   *  \param connection List of vertices
   *  \param label Label for the new edge
   *  \param id Id of the new hyperedge
   *  \return Id of the created hyperedge 
   */
  T addEdge(std::list<T> connection, L label, T id);
  
  //! Find a subgraph matching h in *this 
  /*! Result is a mapping from (vertices UNION edged) of h to
   *  (vertices UNION edges) of this. The mapping is defined for all
   *  edges and for those vertices connected to some edge in h (i.e.
   *  the mapping is total if h is a left hand side of a simple 
   *  transformation rule).
   *  Everytime a matching is found, handler.useThisMatch() is called.
   *  If useThisMatch returns true, findMatching returns that matching.
   *  Otherwise the search continues. If no matching was found,
   *  an empty map is returned.
   *  \param h
   *  \param handler
   *  \return
   */
  std::map<T,T> findMatching(Hypergraph<T,L>& h, MatchHandler<T>& handler);
      
  //! Merges h into *this.
  /*! Attention ! PRE: vertices and edges of h and *this must be disjunct.
   *  \param h
   */
  void uniongraphs(const Hypergraph<T,L>& h);
      
  //! returns the Label of edge e
  L getLabel(T edge) const;

  //! returns the set of vertices
  std::set<T> getVertices() const;

  //! returns the set of edges
  std::set<T> getEdges() const;

  //! returns all the edges whose label matches a given label
  /*! \param label Label of the searched edges
   *  \return All the edges that match label
   */
  std::set<T> getEdges(L label) const;

  //! returns the connected vertices of edge e
  std::list<T> getConnectedVertices(T edge) const;
      
  //! returns the connected edges of vertex v
  std::multiset<T> getConnectedEdges(T vertex) const;

  //! friend method operator<<
  friend std::ostream& operator<<<>(std::ostream& os, const Hypergraph<T,L>& hg);

 private:
  //! Backtracking function
  /*! backtracking function to find a subgraph matching
   *  returns true if a matching was found
   */
  bool findNextMatching(Hypergraph<T,L> &h,
			std::stack<typename std::set<T>::iterator> &mappings,
			std::stack<T> &unmapped_edges,
			std::map<T,T> &phi_v,
			std::map<T,T> &phi_e,
			MatchHandler<T>& matchhandler);

  //! create a new Id that is not used yet
  T createNewId();

  //! comparison between two char*
  bool strEqu(char*,char*) const;
  
  // Hypergraph datastructure:
  std::set<T> vertices; //!< set of vertices
  std::set<T> edges;    //!< set of edges
  
  std::map<T,std::list<T> > connectedV;     //!< vertices connected to an edge
  std::map<T,std::multiset<T> > connectedE; //!< reverse mapping of connected vertex

  std::map<T,L> labels;  //!< labeling function
  // the labelling function does not really work.
  // a comparison function is needed: see file:///doc/stl-manual/html/Multimap.html
  std::multimap<L,T> labelToEdge; //!< reverse labeling function

  T lastId; //!< last Id used
};


//! virtual handler for found matches
template<typename T>
class MatchHandler
{
 public:
  virtual bool useThisMatch(std::map<T,T> match) = 0;
};

#endif // HYPERGRAPH_H
