/*****************************************************************************
 * mincover.h: minimal cover of the paths matching a regular expression
 *****************************************************************************
 * Copyright (C) 2004 Universitaet Stuttgart
 * $Id: mincover.h,v 1.1.2.1 2004/08/30 15:18:23 koziouvi Exp $
 *
 * Authors: Nicolas Relange <relangns@studi.informatik.uni-stuttgart.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************/

#ifndef MINCOVER_H
#define MINCOVER_H

#include <iostream>
#include <map>

template<typename T,typename L>
class Hypergraph;

template<typename T>
class Path;

// forward declarations (see below)
template<typename T,typename L>
class MinCover;

template<typename T,typename L>
std::ostream& operator<< (std::ostream& os, const MinCover<T,L>& mc);

//! This class represents the minimal cover of the paths matching a regular expression
/*! An object of this class is the set of all the smallest paths
 *   matching a regular expression.
 */
template<typename T, typename L>
class MinCover
{
 public:
  //! Constructor
  /*! Create a MinCover object with an empty hypergraph and an empty structure
   */
  MinCover();

  //! Create a minimal cover with paths matching the regular expression: \f$r = \epsilon\f$
  /*! \f$*this = f_G(\epsilon)\f$
   *  \param hg Pointer to the hypergraph to analyse
   */
  MinCover(Hypergraph<T,L> *hg);

  //! Create a minimal cover with paths matching the regular expression: \f$r = l\f$
  /*! \f$*this = f_G(l)\f$
   *  Mode to analyze the hyperdges:
   *  - 0: from vertice 0 to all vertices >= 1
   *  - 1: from all vertices i to vertices j > i
   *  The hyperdges with only one vertice are always analyze in the same way.
   *  \param l \f$r = l\f$ Regular expression all the paths match
   *  \param hg Pointer to the hypergraph to analyse
   *  \param mode Mode to analyze the hyperedges
   */
  MinCover(L l, Hypergraph<T,L> *hg, unsigned mode = 0);

  //! Create a minimal cover with paths matching a regular expression and the constraint in the hyperedges that the label is read from the vertice source to the vertice target
  /*! \f$*this = f_G(l)\f$
   *  \param l \f$r = l\f$ Regular expression all the paths match
   *  \param source In an hyperedge the position of the source vertice \f$source >= 0\f$
   *  \param target In an hyperedge the position of the target vertice \f$target >= 0\f$
   *  \param hg Pointer to the hypergraph to analyse
   *  \see MinCover(L,Hypergraph<T,L>)
   */
  MinCover(L l, const unsigned source, const unsigned target, Hypergraph<T,L> *hg);

  //! Copy constructor
  MinCover(const MinCover<T,L>& rhs);

  //! Destructor
  ~MinCover();

  //! Union
  /*! \f$ *this = f_G(r_1)\f$: the minimal cover matching the regular expression \f$r_1\f$
   *  \param rhs \f$f_G(r_2)\f$: the minimal cover matching the regular expression \f$r_2\f$
   *  \return \f$f_G(r_1+r_2)\f$The minimal cover matching the regular expression \f$r_1+r_2\f$
   *  \sa MinCover::operator+=()
   */
  MinCover<T,L> operator+(const MinCover<T,L> &rhs) const;

  //! Union and assign
  /*! Before the call of the function, \f$this =  f_G(r_1)\f$.
   *  After that, \f$this = f_G(r_1 + r_2)\f$.
   *  \param rhs \f$f_G(r_2)\f$: the minimal cover matching the regular expression \f$r_2\f$
   *  \sa MinCover::operator+()
   */
  void operator+=(const MinCover<T,L> &rhs);

  //! Concatenation
  /*! \f$ this = f_G(r_1)\f$: the minimal cover matching the regular expression \f$r_1\f$
   *  \param rhs \f$f_G(r_2)\f$: the minimal cover matching the regular expression \f$r_2\f$
   *  \return \f$f_G(r_1r_2)\f$The minimal cover matching the regular expression \f$r_1r_2\f$
   *  \sa MinCover::operator*=()
   */
  MinCover<T,L> operator*(const MinCover<T,L> &rhs) const;

  //! Concatenation and assign
  /*! Before the call of the function, \f$this =  f_G(r_1)\f$.
   *  After that, \f$this = f_G(r_1r_2)\f$.
   *  \param rhs \f$f_G(r_2)\f$: the minimal cover matching the regular expression \f$r_2\f$
   *  \sa MinCover::operator*()
   */
  void operator*=(const MinCover<T,L> &rhs);

  //! Kleen closure
  /*! \f$ this = f_G(r)\f$: the minimal cover matching the regular expression \f$r_1\f$
   *  \return \f$f_G(r*)\f$The minimal cover matching the regular expression \f$r*\f$
   *  \sa MinCover::starAss()
   */
  MinCover<T,L> star() const;
  
  //! Kleen closure and assign
  /*! Before the call of the function, \f$this =  f_G(r)\f$.
   *  After that, \f$this = f_G(r*)\f$.
   *  \sa MinCover::star()
   */
  void starAss();

  //! Tests two MinCover for equality
  /*! Tests the following:
   *  - the attributes graph point to the same hypergraph
   *  - the attributes cover are equal
   */
  bool operator==(const MinCover<T,L> &rhs) const;

  //! Tests two MinCover for inequality
  bool operator!=(const MinCover<T,L> &rhs) const;

  //! Add the CTL formula asked by LoLA in a stream
  /*! \param os Stream where the CTL formula must be added. For example, std::cout or std::cerr
   *  \return The modified stream
   */
  std::ostream& lola(std::ostream& os);

  //! Display in the standard output the CTL formula asked by LoLA
  std::ostream& lola();

  //! Add a marking that could be covered by the Petri Net in a stream
  /*! \param os Stream where the marking should be added. For example, std::cout or std::cerr
   *  \return The modified stream
   */
  std::ostream& markings(std::ostream& os);

  //! Display in the standard output the list of markings that could be covered by the Petri Net
  std::ostream& markings();

  //! friend method operator<<
  friend std::ostream& operator<<<>(std::ostream& os, const MinCover<T,L>& mc);

 private:
  //! structure of the minimal cover
  /*! This structures maps a node to every path
   *  that begin with this node.
   */
  std::multimap< T, Path<T> > cover;

  //! pointer to the hypergraph
  Hypergraph<T,L> *graph;

  //! minimize this structure
  /*! If two paths can be ordered,
   *  only the shortest one is kept
   */
  void minimize();

  //! test if a cover is in another one
  /*! \return True, if all the paths of \em this are also in \em rhs
   */
  bool operator%(const MinCover<T,L> &rhs) const;
};

#endif // MINCOVER_H
