/*****************************************************************************
 * path.h: main properties and operations of a path matching a reg exp
 *****************************************************************************
 * Copyright (C) 2004 Universitaet Stuttgart
 * $Id: path.h,v 1.1.2.1 2004/08/30 15:18:23 koziouvi Exp $
 *
 * Authors: Nicolas Relange <relangns@studi.informatik.uni-stuttgart.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************/

#ifndef PATH_H
#define PATH_H

#include <iostream>
#include <map>

// forward declarations (see below)
template<typename T>
struct Path;

template<typename T>
std::ostream& operator<< (std::ostream& os, const Path<T>& p);

//! This structure contains the important characteristics of a path
/*! The characteristic that are useful are:
 *  - the initial node
 *  - the last node
 *  - a map that gives the number of times an edge is gone through
 */
template<typename T>
struct Path
{
  //! initial node
  T begin;
  //! last node
  T end;
  //! gives the number of times an edge is gone through
  std::map<const T,unsigned> times;

  //! comparison operator \f$ \leq \f$
  /*! \f$p_i = (u_i,v_i,f_i), i \in \{1,2\}\f$
   *  \f[ p_1 \leq p_2 \Leftrightarrow \left\{ \begin{array}{l}
   *      u_1 = u_2 \\
   *      v_1 = v_2 \\
   *      (\forall e \in E) f_1(e) \leq f_2(e)
   *      \end{array} \right.
   *  \f]
   */
  bool operator<=(const Path<T> &rhs) const;

  //! comparison operator \f$ \geq \f$
  /*! \f$p_i = (u_i,v_i,f_i), i \in \{1,2\}\f$
   *  \f[ p_1 \geq p_2 \Leftrightarrow \left\{ \begin{array}{l}
   *      u_1 = u_2 \\
   *      v_1 = v_2 \\
   *      (\forall e \in E) f_1(e) \geq f_2(e)
   *      \end{array} \right.
   *  \f]
   */
  bool operator>=(const Path<T> &rhs) const;

  //! comparison operator \f$ = \f$
  /*! \f$p_i = (u_i,v_i,f_i), i \in \{1,2\}\f$
   *  \f[ p_1 = p_2 \Leftrightarrow \left\{ \begin{array}{l}
   *      u_1 = u_2 \\
   *      v_1 = v_2 \\
   *      (\forall e \in E) f_1(e) = f_2(e)
   *      \end{array} \right.
   *  \f]
   */
  bool operator==(const Path<T> &rhs) const;

  //! comparison operator \f$ \neq \f$
  bool operator!=(const Path<T> &rhs) const;

  //! comparison operator \f$ < \f$
  /*! \f[ p_1 < p_2 \Leftrightarrow \left\{ \begin{array}{l}
   *      p_1 \leq p_2 \\
   *      p_1 \neq p_2
   *      \end{array} \right.
   *  \f]
   */
  bool operator<(const Path<T> &rhs) const;

  //! comparison operator \f$ > \f$
  /*! \f[ p_1 > p_2 \Leftrightarrow \left\{ \begin{array}{l}
   *      p_1 \geq p_2 \\
   *      p_1 \neq p_2
   *      \end{array} \right.
   *  \f]
   */
  bool operator>(const Path<T> &rhs) const;

  //! add 2 paths
  /*! \f$p_i = (u_i,v_i,f_i), i \in \{1,2,3\}\f$
   *  \f[ p_3 = p_1 + p_2 \Leftrightarrow \left\{ \begin{array}{l}
   *      u_3 = u_1 \\
   *      v_3 = v_2 \\
   *      v_1 = u_2 \\
   *      f_3 = f_1 + f_2 : e \mapsto f_1(e) + f_2(e)
   *      \end{array} \right.
   *  \f]
   *  If \f$v_1 \neq u_2\f$, an assert stops the program
   *   and displays an error message.
   */
  Path<T> operator+(const Path<T> &rhs) const;

  //! add and assign
  /*! \sa Path::operator+()
   */
  void operator+=(const Path<T> &rhs);

  //! friend method operator<<
  friend std::ostream& operator<<<>(std::ostream& os, const Path<T>& p);
};

#endif // PATH_H
