/*****************************************************************************
 * mincover.cc: minimal cover of the paths matching a regular expression
 *****************************************************************************
 * Copyright (C) 2004 Universitaet Stuttgart
 * $Id: mincover.cc,v 1.1.2.2 2005/02/14 16:53:54 koziouvi Exp $
 *
 * Authors: Nicolas Relange <relangns@studi.informatik.uni-stuttgart.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************/

#ifndef MINCOVER_CC
#define MINCOVER_CC

#if HAVE_CONFIG_H
#  include "config.h"
#endif

#include "common.h"
#include "mincover.h"
#include "error.h"
#include "../hypergraph.h"
#include "path.cc"

#include <assert.h>
#include <set>

extern std::ofstream *of_global;

template<typename T,typename L>
MinCover<T,L>::MinCover()
  : cover(std::multimap<T,Path<T> >()),
    graph(NULL)
{
  Error::info("new MinCover()",(*this),3);
}


template<typename T,typename L>
MinCover<T,L>::MinCover(HYPERGRAPH *hg)
  : cover(std::multimap<T,Path<T> >()),
    graph(hg)
{
  map<IDTYPE, VERTEX> vertices = hg->getVertices();
  typename map<IDTYPE, VERTEX>::const_iterator i;
  for(i = vertices.begin(); i != vertices.end(); ++i)
    {
      Path<T> p;
      p.begin = (i->first);
      p.end = (i->first);
      this->cover.insert(std::pair<const T,Path<T> >(i->first,p));
    }
  Error::info("new MinCover(hypergraph)",(*this),3);
}


template<typename T,typename L>
MinCover<T,L>::MinCover(const MinCover<T,L>& rhs)
  : cover(rhs.cover),
    graph(rhs.graph)
{
  // strange ! There is maybe an error in this copy constructor
  Error::info("new MinCover (copy)",(*this),3);
}


template<typename T,typename L>
MinCover<T,L>::MinCover(L label, HYPERGRAPH *hg, unsigned mode)
  : cover(std::multimap<T,Path<T> >()),
    graph(hg)
{
  typename map<IDTYPE, EDGE>::const_iterator i;
  assert(mode == 0 || mode == 1);
  map<IDTYPE, EDGE> edges = hg->getEdges();
  for(i = edges.begin(); i != edges.end();++i)
    {
      const std::list<T> cv = hg->getConnectedVertices(i->first); //list of connected vertices
      // hyperedge with one vertice
      if(cv.size() == 1)
	{
	  Path<T> p;
	  p.begin = *cv.begin();
	  p.end = p.begin;
	  p.times.insert(std::pair<const T,unsigned>(i->first,1));
	  this->cover.insert(std::pair<const T,Path<T> >(p.begin,p));
	  continue;
	}
      // hyperedge with more than one vertice
      switch(mode)
	{
	case 0: // mode 0
	  {
	    typename std::list<T>::const_iterator j = cv.begin();
	    ++j;
	    for(;j != cv.end();++j)
	      {
		Path<T> p;
		p.begin = *cv.begin();
		p.end = (*j);
		p.times.insert(std::pair<const T,unsigned>(i->first,1));
		this->cover.insert(std::pair<const T,Path<T> >(p.begin,p));
	      }
	  }
	  break;
	case 1: // mode 1
	  {
	    // nj and nk store the position of the iterators j and k in the list
	    unsigned nj,nk;
	    nj = 1;
	    nk = cv.size();
	    for(typename std::list<T>::const_iterator j = cv.begin();
		j != cv.end();
		++j, ++nj, nk = cv.size())
	      for(typename std::list<T>::const_reverse_iterator k = cv.rbegin();
		  k != cv.rend() && nj < nk;
		  ++k, --nk)
		{
		  Path<T> p;
		  p.begin = *j;
		  p.end = *k;
		  p.times.insert(std::pair<const T,unsigned>(i->first,1));
		  this->cover.insert(std::pair<const T,Path<T> >(p.begin,p));
		}
	  }
	  break;
	default: // should raise an error
	  break;
	}
    }
  Error::info("new MinCover(label,hypergraph)",(*this),3);
}


template<typename T,typename L>
MinCover<T,L>::MinCover(L label, const unsigned source, const unsigned target, HYPERGRAPH *hg)
  : cover(std::multimap<T,Path<T> >()),
    graph(hg)
{
  map<IDTYPE, EDGE> edges_ = hg->getEdges();
  map<IDTYPE, EDGE> edges;
  for(map<IDTYPE, EDGE>::iterator it=edges_.begin(); it!=edges_.end(); it++)
  {
	if(it->second.getLabel()==string((const char*)label))
	 edges[it->first] = it->second;
  }
  for(typename map<IDTYPE, EDGE>::const_iterator i = edges.begin();
      i != edges.end();
      ++i)
    {
      Path<T> p;
      const std::list<T> cv = hg->getConnectedVertices(i->first);
      unsigned int k = 0;
      typename std::list<T>::const_iterator j;
      for( j = cv.begin();
	   j != cv.end();
	   ++j, ++k)
	{
	  if(k == source)
	    p.begin = (*j);
	  if(k == target)
	    p.end = (*j);
	}
      p.times.insert(std::pair<const T,unsigned>(i->first,1));
      this->cover.insert(std::pair<const T,Path<T> >(p.begin,p));
    }
  Error::info("new MinCover(label,source,target,hypergraph)",(*this),3);
  Error::info("source",source,3);
  Error::info("target",target,3);
}


template<typename T,typename L>
MinCover<T,L>::~MinCover()
{
}


template<typename T,typename L>
MinCover<T,L>
MinCover<T,L>::operator+(const MinCover<T,L> &rhs) const
{
  MinCover<T,L> res = MinCover<T,L>(*this);
  res += rhs;
  Error::info("MinCover::operator+",(*this),3);
  Error::info("+",rhs,3);
  Error::info("=",res,3);
  return res;
}


template<typename T,typename L>
void
MinCover<T,L>::operator+=(const MinCover<T,L> &rhs)
{
  Error::info("MinCover::operator+=",(*this),3);
  typename std::multimap< T, Path<T> >::const_iterator i;
  for(i = rhs.cover.begin(); i != rhs.cover.end(); ++i)
    {
      this->cover.insert(*i);
    }
  this->minimize();
  Error::info("+",rhs,3);
  Error::info("=",(*this),3);
  return;
}


template<typename T,typename L>
MinCover<T,L>
MinCover<T,L>::operator*(const MinCover<T,L> &rhs) const
{
  MinCover<T,L> res = MinCover<T,L>();
  typename std::multimap< T, Path<T> >::const_iterator i,j;
  for(i = this->cover.begin(); i != this->cover.end(); ++i)
    {
      Path<T> p1 = (*i).second;
      std::pair<typename std::multimap<T,Path<T> >::const_iterator,typename std::multimap<T,Path<T> >::const_iterator> range = rhs.cover.equal_range(p1.end);
      for( j = range.first; j != range.second; ++j)
	{
	  Path<T> p3 = p1 + j->second;
	  res.cover.insert( std::pair< T, Path<T> >(p3.begin,p3) );
	}
    }
  res.minimize();
  Error::info("MinCover::operator*",(*this),3);
  Error::info("*",rhs,3);
  Error::info("=",res,3);
  return res;
}


template<typename T,typename L>
void
MinCover<T,L>::operator*=(const MinCover<T,L> &rhs)
{
  MinCover<T,L> mc3 = (*this) * rhs;
  this->cover = mc3.cover;
  return;
}


template<typename T,typename L>
MinCover<T,L>
MinCover<T,L>::star() const
{
  Error::info("MinCover::star()",(*this),3);
  MinCover<T,L> an = MinCover<T,L>();
  an.graph = this->graph;
  MinCover<T,L> an1 = MinCover<T,L>();
  MinCover<T,L> ri;
  for( ri=MinCover<T,L>(this->graph), an1 += ri;
       an != an1;
       ri *= (*this) )
    {
      an.cover = an1.cover;
      an1 += ri;
    }
  Error::info("=",an1,3);
  return an1;
}


template<typename T,typename L>
void
MinCover<T,L>::starAss()
{
  Error::info("MinCover::starAss()",(*this),3);
  MinCover<T,L> an = MinCover<T,L>(this->graph);
  // add epsilon to (*this)
  (*this) += an;
  MinCover<T,L> ri;
  // in the beginning, ri = f(r)
  for( ri=(*this)*(*this);
       an != (*this);
       ri *= (*this) )
    {
      an.cover = this->cover;
      (*this) += ri;
    }
  Error::info("=",*this,3);
  return;
}


template<typename T,typename L>
bool
MinCover<T,L>::operator==(const MinCover<T,L> &rhs) const
{
  return ( this->graph == rhs.graph && this->cover == rhs.cover );
}


template<typename T,typename L>
bool
MinCover<T,L>::operator!=(const MinCover<T,L> &rhs) const
{
  return !( *this == rhs );
}


template<typename T,typename L>
void
MinCover<T,L>::minimize()
{
  Error::info("MinCover::minimize",*this,3);
  typename std::multimap< T, Path<T> >::iterator i,j;
  for(i = this->cover.begin(); i != this->cover.end(); ++i)
    {
      std::pair<typename std::multimap<T,Path<T> >::iterator,typename std::multimap<T,Path<T> >::iterator> range = this->cover.equal_range(i->second.begin);
      for(j = range.first; j != range.second; ++j)
      {
 	if( (j != i) && (j->second >= i->second) )
 	  this->cover.erase(j);
      }
    }
  Error::info("minimize",*this,3);
  return;
}


template<typename T,typename L>
bool
MinCover<T,L>::operator%(const MinCover<T,L> &rhs) const
{
  typename std::multimap< T, Path<T> >::const_iterator i,j;
  for(i = this->cover.begin(); i != this->cover.end(); ++i)
    {
      bool test = false;
      std::pair<typename std::multimap<T,Path<T> >::const_iterator,typename std::multimap<T,Path<T> >::const_iterator> range = rhs.cover.equal_range(i->first);
      for(j = range.first; j != range.second; ++j)
	if(i->second == j->second)
	  {
	    test = true;
	    break;
	  }
      if(!test)
	return false;
    }
  return true;
}


template<typename T,typename L>
std::ostream&
MinCover<T,L>::lola(std::ostream& os)
{
  os << "FORMULA ";
  if( this->cover.empty() )
    os << "FALSE";
  for(typename std::multimap<T,Path<T> >::const_iterator i = this->cover.begin();
      i != this->cover.end();)
    {
      os << "( ";
      if( i->second.times.empty() )
	os << "TRUE";
      for(typename std::map<const T,unsigned>::const_iterator j = i->second.times.begin();
	  j != i->second.times.end();)
	{
	  os << this->graph->getEdge(j->first).getLabel() << "_" << j->first << " >= " << j->second;
	  ++j;
	  if(j != i->second.times.end())
	    os << " AND ";
	}
      os <<  " )";
      i++;
      if(i != this->cover.end())
	os << " OR ";
    }
  os << "\n";
  return os;
}


template<typename T,typename L>
std::ostream&
MinCover<T,L>::lola()
{
  return this->lola(std::cout);
}


template<typename T,typename L>
std::ostream&
MinCover<T,L>::markings(std::ostream& os)
{
  for(typename std::multimap<T,Path<T> >::const_iterator i = this->cover.begin();
      i != this->cover.end();
      ++i)
    {
      *of_global << "MARKING ";
      for(typename std::map<const T,unsigned>::const_iterator j = i->second.times.begin();
	  j != i->second.times.end();)
	{
	  *of_global << this->graph->getEdge(j->first).getLabel() << "_" << j->first << ": " << j->second;
	  ++j;
	  if(j != i->second.times.end())
	    *of_global<< ", ";
	}
      *of_global << ";\n";
    }
  return os;
}


template<typename T,typename L>
std::ostream&
MinCover<T,L>::markings()
{
  return this->markings(std::cout);
}


template<typename T,typename L>
std::ostream&
operator<< (std::ostream& os, const MinCover<T,L>& mc)
{
  os << "MinCover {\n";
  for(typename std::multimap<T, Path<T> >::const_iterator i = mc.cover.begin();
      i != mc.cover.end();
      i++)
    os << i->second << "\n";
  os << "}\n";
  return os;
}


// instantiations of the templates
template class MinCover<idType,labelType>;

#endif // MINCOVER_CC
