/*****************************************************************************
 * path.cc: main properties and operations of the paths matching a reg exp
 *****************************************************************************
 * Copyright (C) 2004 Universitaet Stuttgart
 * $Id: path.cc,v 1.1.2.2 2005/02/14 16:53:54 koziouvi Exp $
 *
 * Authors: Nicolas Relange <relangns@studi.informatik.uni-stuttgart.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************/

#ifndef PATH_CC
#define PATH_CC

#if HAVE_CONFIG_H
#  include "config.h"
#endif

#include "common.h"
#include "path.h"
#include "error.h"
#include "../hypergraph.h"

#include <assert.h>
#include <set>

extern HYPERGRAPH *spHg;

template<typename T>
bool
Path<T>::operator<=(const Path<T> &rhs) const
{
  // verbose
  Error::info("Path::operator<=",*this,3);
  Error::info(">=",rhs,3);

  if(this->begin != rhs.begin || this->end != rhs.end)
    return false;

  typename std::map<const T,unsigned>::const_iterator i,j;
  typename map<IDTYPE, EDGE>::const_iterator k;
  //test on all edges of the hypergraph *spHg
  map<IDTYPE, EDGE> edges = spHg->getEdges();
  for(k = edges.begin(); k != edges.end(); ++k)
    {
      i = this->times.find(k->first);
      j = rhs.times.find(k->first);

      // 'i->second' == 0
      if(i == this->times.end())
	continue;
      
      // 'j->second' == 0
      if(j == rhs.times.end())
	return false;

      // i->second != 0 && j->second != 0
      if(i->second > j->second)
	return false;
    }
  
  Error::info("Result","true",3);
  return true;
}


template<typename T>
bool
Path<T>::operator>=(const Path<T> &rhs) const
{
  // verbose
  Error::info("Path::operator>=",*this,3);
  Error::info(">=",rhs,3);

  if(this->begin != rhs.begin || this->end != rhs.end)
    return false;
  
  typename std::map<const T,unsigned>::const_iterator i,j;
  typename std::map<IDTYPE, EDGE>::const_iterator k;
  // test on all edges of the hypergraph *spHg
  map<IDTYPE, EDGE> edges = spHg->getEdges();
  for(k = edges.begin(); k != edges.end(); ++k)
    {
      Error::info("k",k->first,3);
      i = this->times.find(k->first);
      j = rhs.times.find(k->first);
      
      // 'j->second' == 0
      if(j == rhs.times.end())
	continue;
      
      // 'i->second' == 0
      if(i == this->times.end())
	return false;
      
      // i->second != 0 && j->second != 0
      if(i->second < j->second)
	return false;
    }
 
  Error::info("Result","true",3);
  return true;
}


template<typename T>
bool
Path<T>::operator==(const Path<T> &rhs) const
{
  // to rewrite and make it quicker
  return (*this <= rhs) && (*this >= rhs);
}


template<typename T>
bool
Path<T>::operator!=(const Path<T> &rhs) const
{
  // to rewrite and make it quicker
  return !(*this == rhs);
}


template<typename T>
bool
Path<T>::operator<(const Path<T> &rhs) const
{
  return (*this <= rhs) && (*this != rhs);
}


template<typename T>
bool
Path<T>::operator>(const Path<T> &rhs) const
{
  return (*this >= rhs) && (*this != rhs);
}


template<typename T>
Path<T>
Path<T>::operator+(const Path<T> &rhs) const
{
  assert(this->end == rhs.begin);

  Path<T> p;
  p.begin = this->begin;
  p.end = rhs.end;
  p.times = this->times;

  typename std::map<const T,unsigned>::const_iterator i;
  for(i = rhs.times.begin(); i != rhs.times.end(); ++i)
    {
      p.times[i->first] += i->second;
    }

  return p;
}


template<typename T>
void
Path<T>::operator+=(const Path<T> &rhs)
{
  assert(this->end == rhs.begin);

  this->end = rhs.end;

  typename std::map<const T,unsigned>::const_iterator i;
  for(i = rhs.times.begin(); i != rhs.times.end(); ++i)
    {
      this->times[i->first] += i->second;
    }
}


template<typename T>
std::ostream&
operator<< (std::ostream& os, const Path<T>& p)
{
  os << "Path " << p.begin << " -> " << p.end << "\tf: ";
  typename std::map<const T,unsigned>::const_iterator i;
  for( i = p.times.begin(); i != p.times.end(); ++i)
    os << i->first << " |-> " << i->second << " ";
  return os;
}


// instantiations of the templates
template class Path<idType>;

#endif // PATH_CC
